"""#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# @author: fmaschietto, bcallen95"""

from silx.io.dictdump import dicttoh5
from silx.io.dictdump import load as dictload
from mdigest.core.toolkit import *
from mdigest.core.imports import *

ds_args = {'compression': "gzip",
           'shuffle': True,
           'fletcher32': True}


class MDSdata:
    """
    Save insances from mdigest.DynCorr, mdigest.DihDynCorr, mdigest.KS_Energy,
    mdigest.CMTY for easy access.
    ``[**]`` function structure adapted from https://github.com/melomcr/dynetan
    """
    def __init__(self):
        self.nodes_to_res_dictionary = None
        self.nodes_dih_to_indices_dictionary = None

        # attributes from correlation.DynCorr
        self.nresidues = None
        self.nnodes = None
        self.nodes_idx_array = None
        self.coordinates_allreplicas = None
        self.displacements_allreplicas = None
        self.distances_allreplicas = None
        self.covariances_allreplicas = None
        self.covar_disp_allreplicas = None
        self.lmi_allreplicas = None
        self.gcc_allreplicas = None
        self.dcc_allreplicas = None
        self.pcc_allreplicas = None
        self.atom_group_selstr = None
        self.system_selstr = None
        self.exclusion_matrix_allreplicas = None
        self.eigenvector_centrality_allreplicas = None

        # attributes from dcorrelation.DihDynCorr
        self.dih_labels = None
        self.dih_indices = None
        self.dihedrals_allreplicas = None
        self.disp_from_mean_dih_allreplicas = None
        self.covar_dih_allreplicas = None
        self.dih_gcc_allreplicas = None
        self.dih_dcc_allreplicas = None
        self.dih_pcc_allreplicas = None
        self.dih_eigen_centrality_allreplicas = None

        # attributes from kscorrelation.KS_energy
        self.bb_distances_allrep = None
        self.KS_energies_allrep = None
        self.KS_DA_energies_allrep = None
        self.KS_DA_MI_corr_allrep = None
        self.KS_DA_LMI_corr_allrep = None
        self.KS_cov_allrep = None

        self.is_protein = None
        self.is_proline = None
        self.q1q2 = None

        self.eigvec_centrality_da_allrep = None
        self.eigvec_centrality_don_allrep = None
        self.eigvec_centrality_acc_allrep = None
        self.eigvec_centrality_da_mean_allrep = None
        self.eigvec_centrality_da_indep_sum_allrep = None

        # attributes from network_communities.CMTY

        self.nodes_communities_collect = None
        self.predecessors = None
        self.nxGraphs_dict = None

        self.partitions_collect = None
        self.distances_collect = None
        self.max_distance = None
        self.max_direct_distance = None

        #  attributes not saved to file but are reconstructed from the loaded information
        self.atom_group_selection = None
        self.num_replicas = None

    def save_to_file(self, file_name_root, save_space=False):
        """
        Opens the HDF5 file and stores all data

        Parameters
        ----------
        file_name_root: str,
            file rootname
        save_space: bool,
            if set to True avoid saving to file some very large attributes.
        """

        with h5py.File(file_name_root + "_cache.hf", "w") as f:

            # atomic displacements correlation related attributes (arrays)
            if self.atom_group_selstr is not None:
                f_atom_group_selstr = f.create_dataset("atom_group_selstr",
                                                       shape=np.asarray(self.atom_group_selstr).astype('S100').shape,
                                                       dtype=np.asarray(self.atom_group_selstr).astype('S100').dtype,
                                                       data=np.asarray(self.atom_group_selstr).astype('S100'))
            if self.system_selstr is not None:
                f_system_selstr = f.create_dataset("system_selstr",
                                                   shape=np.asarray(self.system_selstr).astype('S100').shape,
                                                   dtype=np.asarray(self.system_selstr).astype('S100').dtype,
                                                   data=np.asarray(self.system_selstr).astype('S100'))

            if self.nodes_idx_array is not None:
                f_nodes_idx_array = f.create_dataset("nodes_idx_array",
                                                     shape=self.nodes_idx_array.shape,
                                                     dtype=self.nodes_idx_array.dtype,
                                                     data=self.nodes_idx_array)

            # dihedral correlation related attributes (arrays)
            if self.dih_labels is not None:
                self.dih_labels = np.asarray(self.dih_labels, dtype='S10')
                f_dih_labels = f.create_dataset("dih_labels",
                                                shape=self.dih_labels.shape,
                                                dtype=self.dih_labels.dtype,
                                                data=self.dih_labels)

            if self.dih_indices is not None:
                self.dih_indices = np.asarray(self.dih_indices, dtype=int)
                f_dih_indices = f.create_dataset("dih_indices",
                                                 shape=self.dih_indices.shape,
                                                 dtype=self.dih_indices.dtype,
                                                 data=self.dih_indices)

            # KS_energy related attributes (arrays)
            if self.is_protein is not None:
                f_is_protein = f.create_dataset("is_protein",
                                                shape=self.is_protein.shape,
                                                dtype=self.is_protein.dtype,
                                                data=self.is_protein)
            if self.is_proline is not None:
                f_is_proline = f.create_dataset("is_proline",
                                                shape=self.is_proline.shape,
                                                dtype=self.is_proline.dtype,
                                                data=self.is_proline)
            if self.q1q2 is not None:
                f_q1q2 = f.create_dataset("q1q2",
                                          shape=self.q1q2.shape,
                                          dtype=self.q1q2.dtype,
                                          data=self.q1q2)

            #  communities related attributes

            if self.max_direct_distance is not None:
                self.max_direct_distance = np.asarray([self.max_direct_distance])
                f_max_direct_distance = f.create_dataset("max_direct_distance",
                                                         shape=self.max_direct_distance.shape,
                                                         dtype=self.max_direct_distance.dtype,
                                                         data=self.max_direct_distance)

            if self.max_distance is not None:
                self.max_distance = np.asarray([self.max_distance])
                f_max_distance = f.create_dataset("max_distance",
                                                  shape=self.max_distance.shape,
                                                  dtype=self.max_distance.dtype,
                                                  data=self.max_distance)

                # atomic coordinate correlation  related attributes (dictionaries)
        if self.nodes_to_res_dictionary is not None:
            dicttoh5(self.nodes_to_res_dictionary, file_name_root + '_nodes_to_res_dictionary.h5',
                     create_dataset_args=ds_args)

        if self.nodes_dih_to_indices_dictionary is not None:
            dicttoh5(self.nodes_dih_to_indices_dictionary, file_name_root + '_nodes_dih_to_indices_dictionary.h5',
                     create_dataset_args=ds_args)

        if self.coordinates_allreplicas is not None:
            dicttoh5(self.coordinates_allreplicas, file_name_root + '_coordinates_allreplicas.h5',
                     create_dataset_args=ds_args)

        if self.displacements_allreplicas is not None:
            dicttoh5(self.displacements_allreplicas, file_name_root + '_displacements_allreplicas.h5',
                     create_dataset_args=ds_args)

        if self.distances_allreplicas is not None:
            dicttoh5(self.distances_allreplicas, file_name_root + '_distances_allreplicas.h5',
                     create_dataset_args=ds_args)

        if self.covar_disp_allreplicas is not None:
            dicttoh5(self.covar_disp_allreplicas, file_name_root + '_covar_disp_allreplicas.h5',
                     create_dataset_args=ds_args)

        if self.gcc_allreplicas is not None:
            dicttoh5(self.gcc_allreplicas, file_name_root + '_gcc_allreplicas.h5',
                     create_dataset_args=ds_args)

        if self.dcc_allreplicas is not None:
            dicttoh5(self.dcc_allreplicas, file_name_root + '_dcc_allreplicas.h5',
                     create_dataset_args=ds_args)

        if self.pcc_allreplicas is not None:
            dicttoh5(self.pcc_allreplicas, file_name_root + '_pcc_allreplicas.h5',
                     create_dataset_args=ds_args)

        if self.exclusion_matrix_allreplicas is not None:
            dicttoh5(self.exclusion_matrix_allreplicas, file_name_root + '_exclusion_matrix_allreplicas.h5',
                     create_dataset_args=ds_args)

        if self.eigenvector_centrality_allreplicas is not None:
            dicttoh5(self.eigenvector_centrality_allreplicas, file_name_root + '_eigenvector_centrality_allreplicas.h5',
                     create_dataset_args=ds_args)

        # dihedral correlation related attributes (dictionaries)
        if self.dihedrals_allreplicas is not None:
            dicttoh5(self.dihedrals_allreplicas, file_name_root + '_dihedrals_allreplicas.h5',
                     create_dataset_args=ds_args)

        if self.disp_from_mean_dih_allreplicas is not None:
            dicttoh5(self.disp_from_mean_dih_allreplicas, file_name_root + '_disp_from_mean_dih_allreplicas.h5',
                     create_dataset_args=ds_args)

        if self.covar_dih_allreplicas is not None:
            dicttoh5(self.covar_dih_allreplicas, file_name_root + '_covar_dih_allreplicas.h5',
                     create_dataset_args=ds_args)

        if self.dih_dcc_allreplicas is not None:
            dicttoh5(self.dih_dcc_allreplicas, file_name_root + '_dih_dcc_allreplicas.h5', create_dataset_args=ds_args)

        if self.dih_pcc_allreplicas is not None:
            dicttoh5(self.dih_pcc_allreplicas, file_name_root + '_dih_pcc_allreplicas.h5', create_dataset_args=ds_args)

        if self.dih_gcc_allreplicas is not None:
            dicttoh5(self.dih_gcc_allreplicas, file_name_root + '_dih_gcc_allreplicas.h5', create_dataset_args=ds_args)

        if self.dih_eigen_centrality_allreplicas is not None:
            dicttoh5(self.dih_eigen_centrality_allreplicas, file_name_root + '_dih_eigen_centrality_allreplicas.h5',
                     create_dataset_args=ds_args)

        # KS_energy related attributes (dictionaries)
        if self.KS_energies_allrep is not None:
            if not save_space:
                dicttoh5(self.bb_distances_allrep, file_name_root + '_bb_distances_allrep.h5',
                         create_dataset_args=ds_args)
                dicttoh5(self.KS_energies_allrep, file_name_root + '_energies_allrep.h5', create_dataset_args=ds_args)
        if self.KS_DA_energies_allrep is not None:
            dicttoh5(self.KS_DA_energies_allrep, file_name_root + '_DA_energies_allrep.h5', create_dataset_args=ds_args)
        if self.KS_DA_LMI_corr_allrep is not None:
            dicttoh5(self.KS_DA_LMI_corr_allrep, file_name_root + '_DA_LMI_corr_allrep.h5', create_dataset_args=ds_args)
        if self.KS_DA_MI_corr_allrep is not None:
            dicttoh5(self.KS_DA_MI_corr_allrep, file_name_root + '_DA_MI_corr_allrep.h5', create_dataset_args=ds_args)
        if self.KS_cov_allrep is not None:
            dicttoh5(self.KS_cov_allrep, file_name_root + '_cov_allrep.h5', create_dataset_args=ds_args)
        if self.eigvec_centrality_da_allrep is not None:
            dicttoh5(self.eigvec_centrality_da_allrep, file_name_root + '_eigvec_centrality_da_allrep.h5',
                     create_dataset_args=ds_args)
        if self.eigvec_centrality_don_allrep is not None:
            dicttoh5(self.eigvec_centrality_don_allrep, file_name_root + '_eigvec_centrality_don_allrep.h5',
                     create_dataset_args=ds_args)
        if self.eigvec_centrality_acc_allrep is not None:
            dicttoh5(self.eigvec_centrality_acc_allrep, file_name_root + '_eigvec_centrality_acc_allrep.h5',
                     create_dataset_args=ds_args)
        if self.eigvec_centrality_da_mean_allrep is not None:
            dicttoh5(self.eigvec_centrality_da_mean_allrep, file_name_root + '_eigvec_centrality_da_mean_allrep.h5',
                     create_dataset_args=ds_args)
        if self.eigvec_centrality_da_indep_sum_allrep is not None:
            dicttoh5(self.eigvec_centrality_da_indep_sum_allrep,
                     file_name_root + '_eigvec_centrality_da_indep_sum_allrep.h5', create_dataset_args=ds_args)

        # Graphs and communities related attributes
        if self.nxGraphs_dict is not None:
            with open(file_name_root + "_nxGraphs.pickle", 'wb') as outfile:
                pickle.dump(self.nxGraphs_dict, outfile)

        if self.predecessors is not None:
            dicttoh5(self.predecessors, file_name_root + "_predecessors.h5", create_dataset_args=ds_args)

        if self.nodes_communities_collect is not None:
            dicttoh5(self.nodes_communities_collect, file_name_root + "_nodesComm.h5", create_dataset_args=ds_args)

        if self.partitions_collect is not None:
            self.partitions_collect = dict(zip(np.arange(len(self.partitions_collect)), self.partitions_collect))
            dicttoh5(self.partitions_collect, file_name_root + "_partitions_collect.h5", create_dataset_args=ds_args)

        if self.distances_collect is not None:
            self.distances_collect = dict(zip(np.arange(len(self.distances_collect)), self.distances_collect))
            dicttoh5(self.distances_collect, file_name_root + "_distances_collect.h5", create_dataset_args=ds_args)

    def load_from_file(self, file_name_root, save_space=False):
        """
        DESCRIPTION
        reads cached data and loads attributes
        """

        # load arrays
        if file_exists(file_name_root + '_cache.hf'):

            with h5py.File(file_name_root + "_cache.hf", "r") as f:
                print("@>: cached file found: loading ", file_name_root + '_cache.hf')
                for key, val in f.items():
                    print("@>:", key, f[key].dtype, f[key].shape, f[key].size)
                    if f[key].size > 1:

                        # stores value in object
                        setattr(self, key, np.zeros(f[key].shape, dtype=f[key].dtype))
                        f[key].read_direct(getattr(self, key))

                    elif f[key].size == 1:

                        # for *scalar* H5Py Dataset, use empty touple.
                        setattr(self, key, f[key][()])
                    print('    Done loading %s attribute' % key)
        else:
            print("@>: file '%s' does not extist" % (file_name_root + "_cache.hf"))

        if file_exists(file_name_root + '_nodes_to_res_dictionary.h5'):
            print('@>: load %s' % file_name_root + '_nodes_to_res_dictionary.h5')
            setattr(self, 'nodes_to_res_dictionary',
                    dictload(file_name_root + '_nodes_to_res_dictionary.h5', fmat='hdf5'))

        if file_exists(file_name_root + '_nodes_dih_to_indices_dictionary.h5'):
            print('@>: load %s' % file_name_root + '_nodes_dih_to_indices_dictionary.h5')
            setattr(self, 'nodes_dih_to_indices_dictionary',
                    dictload(file_name_root + '_nodes_dih_to_indices_dictionary.h5', fmat='hdf5'))

        # coordinates correlation related attributes
        if file_exists(file_name_root + '_nodes_to_res_dictionary.h5'):
            print('@>: load %s' % file_name_root + '_nodes_to_res_dictionary.h5')
            setattr(self, 'nodes_to_res_dictionary',
                    dictload(file_name_root + '_nodes_to_res_dictionary.h5', fmat='hdf5'))

        if file_exists(file_name_root + '_nodes_dih_to_indices_dictionary.h5'):
            print('@>: load %s' % file_name_root + '_nodes_dih_to_indices_dictionary.h5')
            setattr(self, 'nodes_dih_to_indices_dictionary',
                    dictload(file_name_root + '_nodes_dih_to_indices_dictionary.h5', fmat='hdf5'))

        if file_exists(file_name_root + '_coordinates_allreplicas.h5'):
            print('@>: load %s' % file_name_root + '_coordinates_allreplicas.h5')
            setattr(self, 'coordinates_allreplicas',
                    dictload(file_name_root + '_coordinates_allreplicas.h5', fmat='hdf5'))

        if file_exists(file_name_root + '_displacements_allreplicas.h5'):
            print('@>: load %s' % file_name_root + '_displacements_allreplicash5')
            setattr(self, 'displacements_allreplicas',
                    dictload(file_name_root + '_displacements_allreplicas.h5', fmat='hdf5'))

        if file_exists(file_name_root + '_distances_allreplicas.h5'):
            print('@>: load %s' % file_name_root + '_distances_allreplicas.h5')
            setattr(self, 'distances_allreplicas',
                    dictload(file_name_root + '_distances_allreplicas.h5', fmat='hdf5'))

        if file_exists(file_name_root + '_covar_disp_allreplicas.h5'):
            print('@>: load %s' % file_name_root + '_covar_disp_allreplicas.h5')
            setattr(self, 'covar_disp_allreplicas',
                    dictload(file_name_root + '_covar_disp_allreplicas.h5', fmat='hdf5'))

        if file_exists(file_name_root + '_gcc_allreplicas.h5'):
            print('@>: load %s' % file_name_root + '_gcc_allreplicas.h5')
            setattr(self, 'gcc_allreplicas',
                    dictload(file_name_root + '_gcc_allreplicas.h5', fmat='hdf5'))

        if file_exists(file_name_root + '_pcc_allreplicas.h5'):
            print('@>: load %s' % file_name_root + '_pcc_allreplicas.h5')
            setattr(self, 'pcc_allreplicas',
                    dictload(file_name_root + '_pcc_allreplicas.h5', fmat='hdf5'))

        if file_exists(file_name_root + '_dcc_allreplicas.h5'):
            print('@>: load %s' % file_name_root + '_dcc_allreplicas.h5')
            setattr(self, 'dcc_allreplicas',
                    dictload(file_name_root + '_dcc_allreplicas.h5', fmat='hdf5'))

        if file_exists(file_name_root + '_exclusion_matrix_allreplicas.h5'):
            print('@>: load %s' % file_name_root + '_exclusion_matrix_allreplicas.h5')
            setattr(self, 'exclusion_matrix_allreplicas',
                    dictload(file_name_root + '_exclusion_matrix_allreplicas.h5', fmat='hdf5'))

        if file_exists(file_name_root + '_eigenvector_centrality_allreplicas.h5'):
            print('@>: load %s' % file_name_root + '_eigenvector_centrality_allreplicas.h5')
            setattr(self, 'eigenvector_centrality_allreplicas',
                    dictload(file_name_root + '_eigenvector_centrality_allreplicas.h5', fmat='hdf5'))

        # KS analysis attributes
        if not save_space:
            if file_exists(file_name_root + '_bb_distances_allrep.h5'):
                setattr(self, 'bb_distances_allrep',
                        dictload(file_name_root + '_bb_distances_allrep.h5', fmat='hdf5'))
                setattr(self, 'KS_energies_allrep',
                        dictload(file_name_root + '_energies_allrep.h5', fmat='hdf5'))

        if file_exists(file_name_root + '_DA_energies_allrep.h5'):
            print('@>: load %s' % file_name_root + '_DA_energies_allrep.h5')
            setattr(self, 'KS_DA_energies_allrep',
                    dictload(file_name_root + '_DA_energies_allrep.h5', fmat='hdf5'))

        if file_exists(file_name_root + '_DA_MI_corr_allrep.h5'):
            print('@>: load %s' % file_name_root + '_DA_MI_corr_allrep.h5')
            setattr(self, 'KS_DA_MI_corr_allrep',
                    dictload(file_name_root + '_DA_MI_corr_allrep.h5', fmat='hdf5'))

        if file_exists(file_name_root + '_DA_LMI_corr_allrep.h5'):
            print('@>: load %s' % file_name_root + '_DA_LMI_corr_allrep.h5')
            setattr(self, 'KS_DA_LMI_corr_allrep',
                    dictload(file_name_root + '_DA_LMI_corr_allrep.h5', fmat='hdf5'))

        if file_exists(file_name_root + '_cov_allrep.h5'):
            print('@>: load %s' % file_name_root + '_cov_allrep.h5')
            setattr(self, 'KS_cov_allrep',
                    dictload(file_name_root + '_cov_allrep.h5', fmat='hdf5'))

        if file_exists(file_name_root + '_eigvec_centrality_da_allrep.h5'):
            print('@>: load %s' % file_name_root + '_eigvec_centrality_da_allrep.h5')
            setattr(self, 'eigvec_centrality_da_allrep',
                    dictload(file_name_root + '_eigvec_centrality_da_allrep.h5', fmat='hdf5'))

        if file_exists(file_name_root + '_eigvec_centrality_don_allrep.h5'):
            print('@>: load %s' % file_name_root + '_eigvec_centrality_don_allrep.h5')
            setattr(self, 'eigvec_centrality_don_allrep',
                    dictload(file_name_root + '_eigvec_centrality_don_allrep.h5', fmat='hdf5'))

        if file_exists(file_name_root + '_eigvec_centrality_acc_allrep.h5'):
            print('@>: load %s' % file_name_root + '_eigvec_centrality_acc_allrep.h5')
            setattr(self, 'eigvec_centrality_acc_allrep',
                    dictload(file_name_root + '_eigvec_centrality_acc_allrep.h5', fmat='hdf5'))

        if file_exists(file_name_root + '_eigvec_centrality_da_mean_allrep.h5'):
            print('@>: load %s' % file_name_root + '_eigvec_centrality_da_mean_allrep.h5')
            setattr(self, 'eigvec_centrality_da_mean_allrep',
                    dictload(file_name_root + '_eigvec_centrality_da_mean_allrep.h5', fmat='hdf5'))

        if file_exists(file_name_root + '_eigvec_centrality_da_indep_sum_allrep.h5'):
            print('@>: load %s' % file_name_root + '_eigvec_centrality_da_indep_sum_allrep.h5')
            setattr(self, 'eigvec_centrality_da_indep_sum_allrep',
                    dictload(file_name_root + '_eigvec_centrality_da_indep_sum_allrep.h5', fmat='hdf5'))

        # dihedral correlation analysis attributes
        if file_exists(file_name_root + '_dihedrals_allreplicas.h5'):
            print('@>: load %s' % file_name_root + '_dihedrals_allreplicas.h5')
            setattr(self, 'dihedrals_allreplicas',
                    dictload(file_name_root + '_dihedrals_allreplicas.h5', fmat='hdf5'))

        if file_exists(file_name_root + '_disp_from_mean_dih_allreplicas.h5'):
            print('@>: load %s' % file_name_root + '_disp_from_mean_dih_allreplicas.h5')
            setattr(self, 'disp_from_mean_dih_allreplicas',
                    dictload(file_name_root + '_disp_from_mean_dih_allreplicas.h5', fmat='hdf5'))

        if file_exists(file_name_root + '_covar_dih_allreplicas.h5'):
            print('@>: load %s' % file_name_root + '_covar_dih_allreplicas.h5')
            setattr(self, 'covar_dih_allreplicas',
                    dictload(file_name_root + '_covar_dih_allreplicas.h5', fmat='hdf5'))

        if file_exists(file_name_root + '_dih_gcc_allreplicas.h5'):
            print('@>: load %s' % file_name_root + '_dih_gcc_allreplicas.h5')
            setattr(self, 'dih_gcc_allreplicas',
                    dictload(file_name_root + '_dih_gcc_allreplicas.h5', fmat='hdf5'))

        if file_exists(file_name_root + '_dih_pcc_allreplicas.h5'):
            print('@>: load %s' % file_name_root + '_dih_pcc_allreplicas.h5')
            setattr(self, 'dih_pcc_allreplicas',
                    dictload(file_name_root + '_dih_pcc_allreplicas.h5', fmat='hdf5'))

        if file_exists(file_name_root + '_dih_dcc_allreplicas.h5'):
            print('@>: load %s' % file_name_root + '_dih_dcc_allreplicas.h5')
            setattr(self, 'dih_dcc_allreplicas',
                    dictload(file_name_root + '_dih_dcc_allreplicas.h5', fmat='hdf5'))

        if file_exists(file_name_root + '_dih_eigen_centrality_allreplicas.h5'):
            print('@>: load %s' % file_name_root + '_dih_eigen_centrality_allreplicas.h5')
            setattr(self, 'dih_eigen_centrality_allreplicas',
                    dictload(file_name_root + '_dih_eigen_centrality_allreplicas.h5', fmat='hdf5'))

        # Graphs and communities related attributes

        if file_exists(file_name_root + '_predecessors.h5'):
            print('@>: load %s' % file_name_root + '_predecessors.h5')
            setattr(self, 'predecessors',
                    dictload(file_name_root + '_predecessors.h5', fmat='hdf5'))

        if file_exists(file_name_root + '_nodesComm.h5'):
            print('@>: load %s' % file_name_root + '_nodesComm.h5')
            setattr(self, 'nodes_communities_collect',
                    dictload(file_name_root + '_nodesComm.h5', fmat='hdf5'))

        if file_exists(file_name_root + '_partitions_collect.h5'):
            print('@>: load %s' % file_name_root + '_partitions_collect.h5')
            setattr(self, 'partitions_collect',
                    dictload(file_name_root + '_partitions_collect.h5', fmat='hdf5'))

        if file_exists(file_name_root + '_distances_collect.h5'):
            print('@>: load %s' % file_name_root + '_distances_collect.h5')
            setattr(self, 'distances_collect',
                    dictload(file_name_root + '_distances_collect.h5', fmat='hdf5'))

        if file_exists(file_name_root + "_nxGraphs_dict.pickle"):
            print('@>: load %s' % file_name_root + '_nxGraphs_dict.pickle')
            with open(file_name_root + "_nxGraphs_dict.pickle", 'rb') as infile:
                self.nxGraphs_dict = pickle.load(infile)

        # load other attributes
        try:
            #if self.nodes_to_res_dictionary is not None:
            self.nresidues = len(np.unique(np.asarray(list(self.nodes_to_res_dictionary.values()))))
        except AttributeError:
            self.nresidues = np.asarray([len(v) for k, v in
                                         self.nodes_communities_collect['0']['comm_nodes'].items()]).sum()

        if self.nnodes is not None:
            self.nnodes = len(np.unique(np.asarray(list(self.nodes_to_res_dictionary.values()))))

        try:
            #if self.eigvec_centrality_don_allrep is not None:
            self.num_replicas = len(self.eigvec_centrality_don_allrep)
        except TypeError:
            try:
                self.num_replicas = len(self.eigenvector_centrality_allreplicas)
            except TypeError:
                try:
                    self.num_replicas = len(self.dih_eigen_centrality_allreplicas)
                except TypeError:
                    nreplicas = len(self.nodes_communities_collect)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# @author: fmaschietto, bcallen95

from mdigest.core.imports import *
from mdigest.core import auxiliary


def plot_eigenvalues_distribution(**kwargs):
    """
    Plot the eigenvalues of a given correlation matrix.

    Parameters
    ----------
    kwargs: dict, possible options are:
            - matrix,      np.ndarray correlation matrix
            - eigenvalues, np.ndarray of eigenvalues
            - dim,         int specifying number of eigenvalues to display
            - color,       sns.palette or array of colors
            - ax,          mpl.ax object
            - matrix_type,   str, title for the legend


    if 'matrix' is provided the latter will be diagonalized and the first n eigenvalues shown as a barplot;
    alternatively, use 'eigenvalues'.
    """

    e_vals = 0
    if kwargs.__contains__('matrix'):
        M = kwargs['matrix']
        e_vals, _ = auxiliary.sorted_eig(M)

    elif kwargs.__contains__('eigenvalues'):
        e_vals = kwargs['eigenvalues']
    else:
        print('@> provide eigenvalues or matrix to diagonalize')

    if kwargs.__contains__('dim'):
        dim = kwargs['dim']
    else:
        dim = 10
    if kwargs.__contains__('color'):
        color = kwargs['color']
    else:
        color = sns.color_palette('bright', dim)[0]
    if kwargs.__contains__('label'):
        label = kwargs['label']
    else:
        label = sns.color_palette('')

    if kwargs.__contains__('ax'):
        ax = kwargs['ax']
        ax.bar(np.arange(len(e_vals[0:dim])), e_vals[0:dim], ec='k', alpha=.8, color=color, label=label)
        ax.set_xticks(np.arange(0, dim, int(np.round(dim / 3, 0))), np.arange(0, dim, int(np.round(dim / 3, 0))) + 1)
        ax.set_xlabel('# eigenvalue')
        ax.set_ylabel('magnitude')


    else:
        plt.figure(figsize=(10,5))
        plt.bar(np.arange(len(e_vals[0:dim])), e_vals[0:dim], ec='k', alpha=.8, label=label)
        plt.xticks(np.arange(0, dim, int(np.round(dim/3, 0))), np.arange(0, dim, int(np.round(dim/3, 0)))+1)
        plt.xlabel('# eigenvalue')
        plt.ylabel('magnitude')

    if kwargs.__contains__('matrix_type'):
        plt.title('Eigenvalues of the {} matrix'.format(kwargs['matrix_type']))







"""#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# @author: fmaschietto, bcallen95"""

from mdigest.core.imports import *
import mdigest.core.savedata as sd
import mdigest.core.toolkit as tk
import mdigest.core.auxiliary as aux

import itertools
import copy
from operator import itemgetter

import community
import networkx as nx
# from networkx.algorithms.community.louvain         import louvain_communities as community_louvain
from networkx.algorithms.shortest_paths.dense import reconstruct_path as reconstruct
from networkx.algorithms.community.centrality import girvan_newman
from networkx.algorithms.community.quality import modularity
from networkx.algorithms.shortest_paths.dense import floyd_warshall_predecessor_and_distance as nxFWPD

import networkx.algorithms.community.quality as nxquality
from networkx import eigenvector_centrality_numpy as nxeigencentrality
from networkx import edge_betweenness_centrality as nxbetweenness


class CMTY:


    def __init__(self):
        """
        Description
        -----------
        Class to handle computation of communities from correlated data

        Attributes
        ----------
        self.cmty_data: obj,
            class object,
        self.verbosity: bool,
            set verbosity
        self.threshold: int,
            impose convergence for Girvan-Newman (GN) algorhitm when number of cycles exceeds  threshold
        self.exclusion_matrix: np.ndarray or dict of np.ndarrays,
            exclusion matrix/ces
        self.distance_matrix: np.ndarray or dict of np.ndarrays,
            distance_matrix,
        self.matrix_dict: nested dict,
            data for community andalysis,
                - example: ``matrix_dictionary = OD({ 'entry_1': dih_gcc_1, 'entry_2':gcc_2})``
        self.pruned_dict: nested dict,
            data for community andalysis where the correlation matrices are filtered
                - ``filters_dictionary = OD({ 'entry_1':excl_mat_1, 'entry_2': excl_mat_1})
                - ``mdigest.CMTY.populate_filters(filters_dictionary=filters_dictionary)``
                - ``filters = {'exclusion': True}``
                - ``mdigest.CMTY.assign_filters(filters=filters)``
        self.filter_dict: nested dict,
            each entry in self.matrix_dict is assigned a filter_dict, which decides which kind of filter will applied to that entry
            ``filter_dict[key] = {'exclusion': True, 'filter_by': False}``
        self.nxGraphs_dict: dict,
            ordered dictionary with the same keys as self.matrix_dict and corresponding nx.Graph as values
        self.nxGraphs: list of nx.Graph object,
            list of graphs where the Graphs appear in the same order as in nxGraphs_dict
        self.num_instances: int,
            number of instances, corresponding to the lenght of self.matrix_dict
        self.list2dict: dict,
            dict with integers (range(num_instances)) as keys and self.matrix_dict keys as values
        self.number_of_nodes: int,
            number or nodes
        self.bestQ_collect: dict,
            best modularity
        self.Qlist_collect: list,
            modularity values list
        self.communities_list_collect: list,
            list of communities from GN
        self.reference_rep_comm_list: list,
            reference iteration in the Louvain procedure (the iteration that affords the largest modularity)
        self.reference_idx: int,
            index of reference iteration
        self.num_iterations: int,
            iteration for Louvain algorithm
        self.nodes_communities_collect: dict,
            nodes as keys, assigned community as values
        self.max_betw_dict_collect: dict,
            maximum betweennes edges at each GN iteration are stored as keys, corresponding betweennesses as values
        self.betweenness: dict,
            betweeness in entire system, for each matrix instance (entry in graph list). 
            Number of shortest paths passign through a given edge computed for each
        self.predecessors: dict, 
            predecessors dictionary
        self.partitions_collect: dict,
            dict with community index as nodes and list of nodes belonging to that community as values
        self.distances_collect: nested  dict,
            dicts, keyed by source and target, of predecessors and distances in the shortest path for each entry in graph list,
            stored in a dictionary with entry name as keys.
        self.max_distance: float
            maximum travelled distance between any two nodes
        self.max_direct_distance: float
            maximum direct distance between two nodes
        """

        self.cmty_data = None
        self.verbosity = None
        self.threshold = None
        self.exclusion_matrix = None
        self.distance_matrix = None
        self.matrix_dict = {}
        self.pruned_dict = OrderedDict()
        self.filter_dict = OrderedDict()
        self.nxGraphs_dict = OrderedDict()
        self.nxGraphs = []
        self.num_instances = 0
        self.list2dict = {}

        # Girvan-Newman / Louvain attributes
        self.number_of_nodes = None
        self.bestQ_collect = None
        self.Qlist_collect = None
        self.communities_list_collect = None
        self.reference_rep_comm_list = None
        self.reference_idx = None
        self.num_iterations = None

        self.nodes_communities_collect = {}
        self.max_betw_dict_collect = {}
        self.betweenness = {}
        #
        self.predecessors = {}
        self.partitions_collect = None
        self.distances_collect = None
        self.max_distance = None
        self.max_direct_distance = None

    def save_class(self, file_name_root='../output/cache/community', save_space=True):
        """
        General function to save instances of the CMTY classs to file

        Parameters
        ----------
        file_name_root: str,
            file rootname

        """

        self.cmty_data = sd.MDSdata()
        self.cmty_data.cmty_data = self.cmty_data
        self.cmty_data.nxGraphs_dict = self.nxGraphs_dict
        self.cmty_data.nodes_communities_collect = self.nodes_communities_collect
        self.cmty_data.max_betw_dict_collect = self.max_betw_dict_collect
        # self.cmty_data.betweenness               = self.betweenness
        self.cmty_data.partitions_collect = self.partitions_collect
        self.cmty_data.distances_collect = self.distances_collect
        self.cmty_data.max_distance = self.max_distance
        self.cmty_data.max_direct_distance = self.max_direct_distance
        if not save_space:
            self.cmty_data.predecessors = self.predecessors
        self.cmty_data.save_to_file(file_name_root)

    # ----------------- general functions to load graphs and parameters --------------------------------------#

    def create_matrix_dict(self, matrix_dictionary):
        """
        Populate matrix_dictionary attribute with kwargs

        Parameters
        ----------
        matrix_dictionary: dict,
            dictionary with format of ``{'matrix_label': np.array or class_object.matrix_attribute}`` containing
            matrices to feed into the community pipeline
        """

        self.matrix_dict = OrderedDict(matrix_dictionary)

    def populate_filters(self, filters_dictionary):
        """
        Populate exclusion_matrix and distance_matrix attributes with filter_dictionary.

        Parameters
        ----------
        filters_dictionary: dict,
            dictionary with format ``filters_dictionary={'exclusion_matrix': mat}``;  mat can be either None, np.array or class_object.matrix_attribute or dict
            containing an exclusion matrix for each matrix in self.matrix_dict.
            The list of keys must match those in self.matrix_dict.
            To include filtering by distance, use ``filters_dictionary={'distance_matrix': mat}``, where mat is as np.array or class_object.matrix_attribute or dict
            containing a distance matrix for each entry in self.matrix_dict. The list of keys must match those in self.matrix_dict.


        Examples
        ---------

         - ``filters_dictionary = { 'exclusion_matrix': np.array}`` or  ``filters_dictionary ='distance_matrix': np.array }`` or
         - ``filters_dictionary = { 'exclusion_matrix': {'ca_lmi_rep_0': np.array,'ca_lmi_rep_1': np.array},
                                  'distance_matrix': {'ca_lmi_rep_0': np.array,'ca_lmi_rep_1': np.array }}``.
        keys in exclusion matrix and distance matrix have to match
        """

        if filters_dictionary.__contains__('exclusion_matrix'):
            self.exclusion_matrix = filters_dictionary['exclusion_matrix']

        if filters_dictionary.__contains__('distance_matrix'):
            self.distance_matrix = filters_dictionary['distance_matrix']

    def assign_filters(self, filters=None):
        """

        Parameters
        ----------
        filters: dict,
            dictionary with format of {'exclusion': bool, 'filter_by': bool}, specifying whether to apply
            specified filter (Default is False for all keys.)
        """

        if filters is None:
            self.filter_dict = OrderedDict()

            for key in self.matrix_dict.keys():
                self.filter_dict[key] = {'exclusion': False, 'filter_by': False}
                warnings.warn("Setting default exclusion/filter criteria. (No filter will be applied.)")
        else:
            # instantiate self.filters
            filters = OrderedDict(filters)
            for key in self.matrix_dict.keys():
                self.filter_dict[key] = {'exclusion': False, 'filter_by': False}

                if filters.__contains__('exclusion'):
                    self.filter_dict[key] = {'exclusion': True, 'filter_by': False}

                if filters.__contains__('filter_by'):
                    self.filter_dict[key] = {'exclusion': False, 'filter_by': True}

            if self.exclusion_matrix is None:
                warnings.warn(
                    "No exclusion matrix has been loaded, call self.populate_filters({'exclusion_matrix': exc_matrix}) "
                    "to suppress this warning.")
            if self.distance_matrix is None:
                warnings.warn(
                    "No distance matrix has been loaded, call self.populate_filters({'distance_matrix': dist_matrix}) "
                    "to suppress this warning.")

    def set_parameters(self, parameters):
        """
        Set parameters
        """
        if parameters.__contains__("VERBOSE"):
            self.verbosity = parameters['VERBOSE']

        else:
            self.verbosity = False
        print("@>: verbosity = %s" % self.verbosity)

        if parameters.__contains__("THRESHOLD"):
            self.threshold = parameters['THRESHOLD']
        else:
            self.threshold = None

        if parameters.__contains__("LOUVAIN_ITERATIONS"):
            self.num_iterations = parameters['LOUVAIN_ITERATIONS']
            print("@>: number of iterations is set to %d" % self.num_iterations)
        else:
            self.num_iterations = None
            print("@>: Louvain set to None")

    def load_graph(self, distance_threshold=5.0):
        """
        General function to load graph

        Parameters
        ----------
        distance_threshold: threshold applied in prune_adjacency

        """

        for key, mat_instance in self.matrix_dict.items():
            print('processing %s ...' % key)

            # We substitute zeros for a non-zero value to avoid "zero division" warnings
            #   from the np.log transformation below.
            mat_instance[np.where(mat_instance == 0)] = 10 ** -11

            # Use log transformation for network distance calculations.

            mat_instance = -1.0 * np.emath.log(mat_instance)

            # Now we guarantee that the previous transformation does not
            #   create "negative infitite" distances. We set those to zero.
            mat_instance[np.where(np.isinf(mat_instance))] = 0.

            if isinstance(mat_instance[0, 0], complex):
                dimensions = mat_instance.shape
                cast2real = np.zeros((dimensions[0] ** 2))
                mat_instance = mat_instance.flatten()
                for i in range(len(mat_instance)):
                    cast2real[i] = mat_instance[i].real
                mat_instance = cast2real.reshape(dimensions)

            if self.filter_dict.__contains__(key):
                filter_by = self.filter_dict[key]['filter_by']
                exclusion = self.filter_dict[key]['exclusion']
            else:
                filter_by = self.filter_dict['filter_by']
                exclusion = self.filter_dict['exclusion']

            if filter_by:
                if isinstance(self.distance_matrix, dict):
                    mat_instance = aux.prune_adjacency(mat_instance, self.distance_matrix[key],
                                                       loc_factor=distance_threshold)
            elif exclusion:
                if isinstance(self.exclusion_matrix, dict):
                    mat_instance[self.exclusion_matrix[key] == 1] = 0.
                else:
                    print('@>: loaded {} matrix as nxGraph'.format(key))

                    mat_instance[np.asarray(self.exclusion_matrix, dtype=int) == 1] = 0.
                self.pruned_dict.update({key: mat_instance})
            else:
                pass

            print(f"@>: there are {np.count_nonzero(mat_instance):d} nonzero elements in this matrix")
            self.nxGraphs_dict[key] = aux.create_graph(mat_instance)
            for pair in self.nxGraphs_dict[key].edges.keys():
                self.nxGraphs_dict[key].edges[(pair[0], pair[1])]["dist"] = mat_instance[pair[0], pair[1]]
            print('@>: graph has %d edges' % len(self.nxGraphs_dict[key].edges))

            # Sets the degree of each node.
            degree_dict = dict(self.nxGraphs_dict[key].degree(self.nxGraphs_dict[key].nodes()))
            nx.set_node_attributes(self.nxGraphs_dict[key], degree_dict, 'degree')
            self.number_of_nodes = len(nx.nodes(self.nxGraphs_dict[key]))
            self.nxGraphs.append(self.nxGraphs_dict[key])
            self.num_instances = len(self.nxGraphs)

            self.list2dict = dict(zip(np.arange(len(self.nxGraphs_dict.keys())), list(self.nxGraphs_dict.keys())))

    # ----------------- community partitioning --------------------------------------#

    def _betweenness(self, predecessors, n):
        """
        Calculate betweenness
        """
        betweenmat = np.zeros((n, n))
        for i in predecessors.keys():
            for j in predecessors[i].keys():
                path = reconstruct(i, j, predecessors)
                for k in range(len(path) - 1):
                    betweenmat[path[k], path[k + 1]] += 1
                    betweenmat[path[k + 1], path[k]] += 1
        return betweenmat

    def most_valuable_edge_fw(self, G):
        """
        Returns most vauable edge, using Floyd-Warshall algoritm
        """

        pred, distances = nx.floyd_warshall_predecessor_and_distance(G)
        n = G.number_of_nodes()
        betweenness_mat = self._betweenness(pred, n)
        edge = np.unravel_index(np.argmax(betweenness_mat), np.array(betweenness_mat).shape)
        if self.verbosity:
            print("@>: most valuable edge:", list(edge))
        return (edge[0], edge[1]), np.amax(betweenness_mat)

    def most_valuable_edge(self, G, count_entries=False, normalized=False, weight='weight'):
        """ Returns most valuable edge according to edge_betweenness_centrality criterion

        Parameters
        ----------
        count_entries: bool,
            use true to print betweenness values calculated without averaging over all shortest paths
            this is how betweenness values are calculated in the original floyd_warshall.c code.
        normalized: bool,
            decides whether betweennesses are normalized or not

        weight: str,
            default 'weigth', uses graph weights

        Returns
        --------
        (edge_key[0], edge_key[1]): tuple,
        maxbet: float
            (edge tuple), maximum_betweeenness
        """

        bw = nx.edge_betweenness_centrality(G, normalized=normalized, weight=weight)
        maxbet = max(bw.values())
        edge_key = tk.keywithmaxval(bw)
        if count_entries:
            A = nx.to_numpy_matrix(G, nodelist=None, weight=None)
            np.fill_diagonal(A, 0.)
            modA = np.linalg.norm(A) / 2
            maxbet = max(bw.values()) * modA
        if (self.verbosity == True) and count_entries:
            print('@>: most valuable edge - count entries:', edge_key, maxbet, modA)

        elif (self.verbosity == True) and not count_entries:
            print('@>: most valuable edge:', edge_key, maxbet)
        return (edge_key[0], edge_key[1]), maxbet

    def _communities_to_partitions(self, communities):
        """
        Convert communities to partitions
        """
        partitions = {}
        for cind, comm in enumerate(communities):
            for n, node in enumerate(comm):
                partitions.update({comm[n]: cind})

        partitions = {key: value for key, value in sorted(partitions.items(), key=lambda item: item[0])}
        return partitions

    def community_data(self, G, partition):
        """
        Compact all information on the communities into a dictionary

        Parameters
        ----------

        G: nx.Graph(),
            a networkx protein graph
        partition: dict,
            dictionary containing the different partitions

        Returns
        ---------
        nodes_communities: dict, n_communities: int
            a dictionary containing data relative to each community:
             - community labels,
             - community index ordered by modularity,
             - community index ordered by eigenvector centrality,
             - community nodes (list of nodes in each partition)
        """

        nodes_communities = {}
        communities_labels = np.unique(np.asarray(list(partition.values()), dtype=int))

        n_communities = len(np.unique(list(partition.values())))
        nodes_communities["comm_labels"] = copy.deepcopy(communities_labels)
        nx.set_node_attributes(G, partition, 'modularity')

        nodes_communities["comm_nodes"] = {}

        for comm_ind, comm in enumerate(communities_labels):
            nodes_in_comm = [n for n in G.nodes() if G.nodes[n]['modularity'] == comm]

            # Then create a dictionary of the eigenvector centralities of those nodes
            self._calc_eigen_central(G)
            nodes_in_comm_eigvs = {n: G.nodes[n]['eigenvector'] for n in nodes_in_comm}

            # Then sort that dictionary
            nodes_in_comm_eigvs_ord = sorted(nodes_in_comm_eigvs.items(), key=itemgetter(1), reverse=True)
            nodes_in_comm_eigvs_ordlist = [x[0] for x in nodes_in_comm_eigvs_ord]

            nodes_communities["comm_nodes"][comm] = copy.deepcopy(nodes_in_comm_eigvs_ordlist)

        return nodes_communities, n_communities

    def sort_cmty(self, cycles, setgraph=-1):
        """
        Sort communities and store sorted indices according to different metrics to a dictionary

        Parameters
        ----------
        cycles: int,
            assign the number of cycles to match the number of replicas (number of graphs on which
            to iterate)
        setgraph: int,
            which graph to use;  if -1 use graph corresponding to louvain cycle
        """

        modularity_collect = []
        self.reference_rep_comm_list = []

        for cycle in range(cycles):

            if setgraph == -1:
                G = self.nxGraphs[cycle]
            else:
                G = self.nxGraphs[setgraph]

            # Orders communities based on size.

            communities_by_size = list(sorted(self.nodes_communities_collect[cycle]["comm_nodes"].keys(),
                                              key=lambda k: len(self.nodes_communities_collect[cycle]["comm_nodes"][k]),
                                              reverse=True))

            self.nodes_communities_collect[cycle]["comm_by_size"] = copy.deepcopy(communities_by_size)

            # Orders communities based on highest eigenvector centrality of all its nodes.
            communities_by_eigen_centr = list(sorted(self.nodes_communities_collect[cycle]["comm_nodes"].keys(),
                                                     key=lambda k:
                                                     G.nodes[self.nodes_communities_collect[cycle]["comm_nodes"][k][0]][
                                                         'eigenvector'], reverse=True))

            self.nodes_communities_collect[cycle]["comm_by_eigen_centr"] = copy.deepcopy(communities_by_eigen_centr)

            # Creates a list of repetitions and order them according to graph modularity.

            modul = nxquality.modularity(G, [set(nodes_list) for nodes_list in
                                             self.nodes_communities_collect[cycle]["comm_nodes"].values()])

            modularity_collect.append((cycle, modul))

            modularity_collect.sort(key=lambda x: x[1], reverse=True)

            # Keep the instance with the highest modularity as a reference for community matching
            reference_idx = modularity_collect[0][0]

        self.reference_idx = reference_idx

        print("@>: reference instance", self.reference_idx, "with modularity = ", modularity_collect[0][1])

        for cycle in range(cycles):
            color_index = np.zeros(len(self.nodes_communities_collect[cycle]["comm_nodes"])).astype(int)

            # Assingn color index accoding to max common element
            for c in range(self.communities_list_collect[cycle]):
                comm = [k for k in self.partitions_collect[cycle].keys() if self.partitions_collect[cycle][k] == c]
                color_index[c] = self._max_common_elements(comm, list(
                    self.nodes_communities_collect[self.reference_idx]["comm_nodes"].values()))
            self.nodes_communities_collect[cycle].update({'color_index': color_index})
        self.reference_rep_comm_list = list(self.nodes_communities_collect[self.reference_idx]["comm_nodes"].values())

    def girvannewman(self, MVE=None):
        """
        Computes Girvan Newman algorithm, slightly faster than Run_Girvan_Newman(), uses builtin functions from nx

        Parameters
        ----------
        MVE: function
            function to calculate most valuable edge. Default is None, which is equivalent to calling
            most_valuable_edge_nx().

        """

        Qlist_collect = {}
        max_btw_dict_collect = {}
        bestQ_collect = {}
        nodes_communities_collect = {}
        communities_list_collect = []
        partitions_collect = []

        for g_idx, current_G in enumerate(self.nxGraphs):

            print('@>: run Girvan Newman for graph %d' % g_idx, self.list2dict[g_idx])
            G = current_G
            G_o = current_G.copy()

            bestQ = 0.0
            Q_list = []

            counter = 1
            if self.threshold is None:
                self.threshold = 10
                print("@>: threshold is set to %d" % self.threshold)
            k = self.threshold  # keep this reasonably large, so that the maximum modularity split corresponds to a number of
            # communities < 50
            undivided = [set(G.nodes()), set()]
            Q = modularity(G, undivided)  # weight is default
            Q_list.append(Q)

            comp = girvan_newman(G)

            limited = itertools.takewhile(lambda commu: len(commu) <= k, comp)

            for communities_ in limited:
                counter += 1

                Q = modularity(G, communities_)

                if Q > bestQ:
                    bestQ = Q
                    communities = list(communities_)

                Q_list.append(Q)

                if self.verbosity:
                    if bestQ > 0.0:
                        print("@>: max modularity (Q):", bestQ)
                        print("@>: graph communities:", communities)
                    else:
                        print("@>: max modularity (Q):", bestQ)

            Qlist_collect.update({g_idx: Q_list})
            bestQ_collect.update({g_idx: bestQ})
            communities = [list(comm) for comm in communities]

            nodes_communities_collect[g_idx] = {}
            partition = self._communities_to_partitions(communities)

            nodes_communities, n_communities = self.community_data(G, partition)
            nodes_communities_collect[g_idx] = nodes_communities

            communities_list_collect.append(n_communities)
            partitions_collect.append(partition)

        self.bestQ_collect = bestQ_collect
        self.Qlist_collect = Qlist_collect
        self.nodes_communities_collect = nodes_communities_collect
        self.partitions_collect = partitions_collect
        self.communities_list_collect = communities_list_collect
        self.max_betw_dict_collect = max_btw_dict_collect

        self.sort_cmty(self.num_instances, setgraph=-1)


    def best_iteration_louvain(self, G, nodes_comm_alliter, comm_list_alliter, partitions_alliter):
        """
        Select the best iteration from Louvain heuristic algorithm by selecting
        the instance with the highest modularity

        Parameters
        ----------
        G: nx.Graph(),
            a networkx protein graph
        nodes_comm_alliter: dict,
            dictionary of Dictionary containing the list of nodes for each community for at every iteration of Louvain
            heuristic algorithm
        comm_list_alliter: list,
            contains the list of nodes for each community at every iteration of Louvain heuristic algorithm
        partitions_alliter: list
            partitions at every iteration of Louvain heuristic algorithm

        Returns
        --------
         best_iteration_comm_nodes: dict,
            communities nodes for the best iteration
        best_comm_list: list,
            list containing the index of the communities in the `best run` best_partitions
        best_partitions: dict,
            best partitions
        """

        modularity_alliter = []
        reference_idx = 0
        for itera in range(self.num_iterations):
            # Creates a list of repetitions and order them according to graph modularity.
            modul = nxquality.modularity(G, [set(nodes_list) for nodes_list in
                                             nodes_comm_alliter[itera]["comm_nodes"].values()])

            modularity_alliter.append((itera, modul))
            modularity_alliter.sort(key=lambda x: x[1], reverse=True)

            # Keep the instance with the highest modularity as a reference for community matching
            reference_idx = modularity_alliter[0][0]

        best_iteration_comm_nodes = nodes_comm_alliter[reference_idx]
        best_comm_list = comm_list_alliter[reference_idx]
        best_partitions = partitions_alliter[reference_idx]
        return best_iteration_comm_nodes, best_comm_list, best_partitions

    def _max_common_elements(self, list1, lists_ref):
        """
        Assign community according to maximum overlap with reference communities.
        """
        common = []
        for lr in lists_ref:
            common.append(np.sum([l in lr for l in list1]))
        return np.argmax(common)

    def _calc_eigen_central(self, G):
        """
        Compute eigenvector centrality to rank the nodes in the network.
        """

        cent = nxeigencentrality(G, weight='weight')
        nx.set_node_attributes(G, cent, 'eigenvector')

    def run_cmty_louvain(self, setgraph=0):
        # TODO check if this function is actually necessary (other than run_cmtys_louvain())
        # TODO if keep add aggregate option

        """
        Runs community louvain (one graph) and returns a dictionary with communities at each iteration
        of the louvain procedure, can be useful if one wants to check consistency across different louvain iterations for a
        single replica.

        Parameters
        ----------
        setgraph: int,
            default 0, assing to integer correponding to replica on which to apply Louvain algorithm
        """

        np.random.seed(1)

        G = self.nxGraphs[setgraph]
        print("@>: partitioning graph n. {} in graphs list".format(setgraph))
        print("@>: louvain algorithm will be iterated {} times".format(self.num_iterations))

        self.nodes_communities_collect = {}
        self.partitions_collect = []
        self.communities_list_collect = []
        for itera in range(self.num_iterations + 1):
            self.nodes_communities_collect[itera] = {}

            partition = community.best_partition(G, weight='degree')
            # partition = community_louvain(G, weight="weight", resolution=0, threshold=0.0000001, seed=None)
            # tmp_ = {}
            # partition = [list(p) for p in partition]
            # for i, p in enumerate(partition):
            #     for node in p:
            #         tmp_.update({node:i})
            # partition = tmp_

            nodes_communities, n_communities = self.community_data(G, partition)
            self.nodes_communities_collect[itera] = nodes_communities
            self.communities_list_collect.append(n_communities)
            self.partitions_collect.append(partition)
        self.sort_cmty(self.num_iterations, setgraph=setgraph)

    def run_cmtys_louvain(self, aggregate=False, **kwargs):
        """
        Community generation using the louvain protocol, iterate over multiple replicas, for each save
        the partition with higher modularity to a dictionary

        Parameters
        ----------
        aggregate: bool,
            whether to group communities by redistributing nodes of communities smaller than given threshold
            over the other communities. Aggregation assign each node to the partition that has yields the larges
            modularity.
        kwargs: dict,
            use threshold = int to set threshold for regrouping communities,
            default is 5 (communities <= 5 elements are redistributed)
        """

        def _reassign_and_modularity(graph, partitions, lone_partition_list):
            """Reassign node to the community which affords the largest network modularity
            """
            original = partitions.copy()
            tmp_partitions = original.copy()

            for lp, p in enumerate(lone_partition_list):
                discard = tmp_partitions.pop(p)

                print('@>: reassign node {}'.format(discard))

                modularity_arr = np.zeros(len(original))

                for pidx, (key, val) in enumerate(tmp_partitions.items()):
                    tmp_partitions[key] += discard

                    modul = nxquality.modularity(graph, [list(nodes_list) for nodes_list in tmp_partitions.values()])
                    modularity_arr[key] = modul

                    tmp_partitions[key] = tmp_partitions[key][:-len(discard)]
                    best_modularity = np.argmax(modularity_arr)

                tmp_partitions[best_modularity] += discard

            final_partitions = {}
            for i, (key, val) in enumerate(tmp_partitions.items()):
                final_partitions.update({i: val})

            return final_partitions

        np.random.seed(1)
        self.nodes_communities_collect = {}
        self.communities_list_collect = []
        self.partitions_collect = []

        for instance in range(self.num_instances):
            np.random.seed(1)
            G = self.nxGraphs[instance]

            communities_list_alliter = []
            partitions_alliter = []

            nodes_communities_alliter = {}
            for itera in range(self.num_iterations):
                partition = community.best_partition(G, weight='degree')

                ###############
                # uncomment to use community louvain package
                res = {}
                for k, v in partition.items():
                    res[v] = [k] if v not in res.keys() else res[v] + [k]
                partition = tk.dict2list(res)

                ###############
                # uncomment to use nx implementation of community_louvain
                # partition = community_louvain(G, weight="weight", resolution=0, threshold=0.0000001, seed=None)
                # partition = [list(p) for p in partition]
                ###################################

                if aggregate:
                    if kwargs:
                        thr = kwargs['threshold']
                    else:
                        thr = 5
                    ref_partitions = tk.list2dict(partition)
                    keys = []
                    for k, v in ref_partitions.items():
                        if 0 < len(v) <= thr:
                            keys.append(k)

                    partition = _reassign_and_modularity(G, ref_partitions, keys)
                else:
                    partition = tk.list2dict(partition)
                tmp_ = {}
                for k, part in partition.items():
                    for node in part:
                        tmp_.update({node: k})
                partition = tmp_

                nodes_communities, n_communities = self.community_data(G, partition)
                nodes_communities_alliter.update({itera: nodes_communities})
                communities_list_alliter.append(n_communities)
                partitions_alliter.append(partition)

            best_iter_comm_nodes, best_comm_list, best_partitions = self.best_iteration_louvain(G,
                                                nodes_communities_alliter, communities_list_alliter, partitions_alliter)

            self.nodes_communities_collect.update({instance: best_iter_comm_nodes})
            self.communities_list_collect.append(best_comm_list)
            self.partitions_collect.append(best_partitions)

        self.sort_cmty(self.num_instances, setgraph=-1)
        if self.verbosity:
            print('@>: partitions_collect', len(self.partitions_collect))
            print('@>: communities_list_collect', len(self.communities_list_collect))
            print('@>: communities keys', self.nodes_communities_collect.keys())

    # -------------------- community analysis --------------------------------------

    def _inter_community_betweeness(self, partition, G):
        """
        Calculate inter community betweenness

        Parameters
        ----------
        partition: dict,
        G: nx.Graph

        Returns
        -------
        :returns intercomm: dict,
            inter community betweenness
        """
        B = nx.edge_betweenness_centrality(G)
        n_communities = len(np.unique(list(partition.values())))
        intercomm = np.zeros((n_communities, n_communities))
        for key in B.keys():
            c1 = partition[key[0]]
            c2 = partition[key[1]]
            intercomm[c1, c2] += B[key]
            intercomm[c2, c1] += B[key]
        np.fill_diagonal(intercomm, 0)
        return intercomm

    def calculate_betweenness(self):
        """
        Calculate betweenness in entire system, for each matrix instance (entry in graph list).
        Store each in an ordered dictionary
        """
        betweenness = {}

        for instance in tk.log_progress(range(self.num_instances), every=1, size=self.num_instances, name="Instance"):
            # Compute betweeness in entire system, for each matrix instance (entry in graph list).

            # neglect weights - the betweenness estimates the number of shortest paths passign through a given edge
            betweenness[instance] = nxbetweenness(self.nxGraphs[instance], weight=None)

            # Creates an ordered dict of pairs with betweenness > than zero.
            betweenness[instance] = {k: self.betweenness[instance][k] for k in self.betweenness[instance].keys() if
                                          self.betweenness[instance][k] > 0}
            self.betweenness[instance] = OrderedDict(
                sorted(betweenness[instance].items(), key=lambda t: t[1], reverse=True))

    def get_degree(self, instance=0):
        return dict(self.nxGraphs[instance].degree(self.nxGraphs[instance].nodes()))

    def compute_optimal_paths(self):
        """
        Compute optimal source-target paths using the Floyd-Warshall algorithm
        """

        # get the network distance arrays
        self.distances_collect = np.zeros([self.num_instances, self.number_of_nodes, self.number_of_nodes],
                                          dtype=np.float)
        self.predecessors = {}
        for i in range(self.num_instances):
            self.predecessors[i] = 0

        for instance in tk.log_progress(range(self.num_instances), name="Instance"):
            # use the "distance" as weight, (i.e. the log-transformation of the correlations, rather than
            # the correlation itself.
            fw_predecessors, fw_distances = nx.floyd_warshall_predecessor_and_distance(self.nxGraphs[instance],
                                                                                       weight='dist')
            # turns dictionary of distances into 2D np.array per instance
            self.distances_collect[instance, :, :] = np.array(
                [[fw_distances[i][j] for i in sorted(fw_distances[j])] for j in sorted(fw_distances)])

            # combines predecessor dictionaries from all replicas and save in class attribute
            self.predecessors[instance] = copy.deepcopy(fw_predecessors)

        # get maximum network distance
        pruned = np.asarray(list(self.pruned_dict.values()))
        self.max_distance = np.max(pruned[self.distances_collect != np.inf])

        # set -1 as distance of nodes with no connecting path (rather than np.inf)
        self.distances_collect[np.where(np.isinf(self.distances_collect))] = -1

        # check max network distance between directly connected nodes
        self.max_direct_distance = max(
            [self.distances_collect[instance, pruned[instance, :, :] > 0].max() for instance in
             range(self.num_instances)])

    def retrieve_path(self, node_A, node_B, instance):
        """
        Reconstruct path
        """
        return nx.reconstruct_path(node_A, node_B, self.predecessors[instance])





def display_shortes_path(nvView, path, dists, max_direct_dist, selected_atomnodes, opacity=0.75,
                         color="green", side="both", segments=5, disable_impostor=True, use_cylinder=True):
    """
    Display shortest paths
    """
    for source, target in zip(path, path[1:]):

        source_sel = tk.get_NGLselection_from_node(source, selected_atomnodes)
        target_sel = tk.get_NGLselection_from_node(target, selected_atomnodes)

        if dists[source, target] == 0:
            continue

        radius = 0.05 + (0.5 * (dists[source, target] / max_direct_dist))

        nvView.add_representation("distance", atom_pair=[[source_sel, target_sel]], color=color, label_visible=False,
                                  side=side, name="link", use_cylinder=use_cylinder, radial_sements=segments,
                                  radius=radius, disable_impostor=disable_impostor, opacity=opacity, lazy=True)




# create universe class
from mdigest.utils.imports import *
from mdigest.utils import toolkit as tk

from MDAnalysis.analysis import align as mdaAlign
from MDAnalysis.analysis import rms
import mdtraj as md

import pickle as pkl
import os
from numba import jit, prange
from tqdm.notebook import trange


# global functions
def pkl_dump(todump, filename):
    with open(filename, 'wb') as handle:
        pkl.dump(todump, handle, protocol=pkl.HIGHEST_PROTOCOL)
    handle.close()
    return


class MDAUniverse:
    """General class to handle and align the MDA Universe
    """
    def __init__(self):
        self.traj = None
        self.topo = None

        self.traj_ref = None
        self.topo_ref = None

        self.mda_u = None
        self.mda_ref_u = None

        self.selection_rmsd = None           # indices of atoms to use for rmsd calculation
        self.selection = None                # select indeces of a subset of the topology of interest
        self.selection_alignment = None      # the indices of the atoms to superpose (mobile)
        self.selection_alignment_ref = None  # the indices of the atoms to superpose (reference)

        self.rmsd = None
        self.name = None
        self.box = None
        self.translate = None


    def set_name(self, name):
        self.name = name

    def load_universe(self, topo, traj, topo_ref, traj_ref):
        self.mda_u = mda.Universe(topo, traj)
        self.mda_ref_u = mda.Universe(topo_ref, traj_ref)

    def source_universe(self, universe, reference=None):
        self.mda_u = universe.mda_u
        if reference is not None:
            self.mda_ref_u = reference
        else:
            self.mda_ref_u = universe.mda_u

    def set_selection(self, atm_selection, segIDs):
        """
        Parameters
        ----------
        atm_selection: str
        Atom selection using MDAnalysis selection string

        segIDs: str
        which segids to select

        Example selection : 'not resid 455'
        Example segIDs : '*'
        """
        self.selection_alignment = "segid " + " ".join(segIDs) + " and not (name H* or name [123]H*) and " + atm_selection

    def set_selection_ref(self, atm_selection, segIDs):
        """ Example selection : 'not resid 455'
            Example segIDs : '*'
        """
        self.selection_alignment_ref = "segid " + " ".join(segIDs) + " and not (name H* or name [123]H*) and " + atm_selection

    def set_selection_rmsd(self, selection):
        self.selection_rmsd = selection


    def align_traj(self, inmem=True, aligntoavg=False, strict=True, center=False):
        """
        Use MDAnalysis to align trajectory with respect to a reference

        This function assumes that atoms in reference traj and mobile traj are in the same order.
        if not use align_mdtraj() which allows to select different atoms in mobile and reference.

        Parameters
        ----------
        inmem: bool,
        aligntoavg: bool,
        strict: bool,
        center: bool,

        Returns
        -------
        """

        mda_u = self.mda_u
        mda_ref_u = self.mda_ref_u

        # Set the first frame as reference for alignment
        self.mda_u.trajectory[0]
        self.mda_ref_u.trajectory[0]

        if aligntoavg:
            # Set the average coordinates as reference for alignment
            atomgroup_ref = self.mda_ref_u.select_atoms(self.selection_alignment_ref)
            reference_coordinates = self.mda_ref_u.trajectory.timeseries(asel=atomgroup_ref).mean(axis=1)
            reference = mda.Merge(atomgroup_ref).load_new(reference_coordinates[None, :, :], order="fac")
            mdaAlign.AlignTraj(mda_u, reference, select=self.selection_alignment, in_memory=inmem, weights="mass",
                               strict=strict).run()
        else:
            mdaAlign.AlignTraj(mda_u, mda_ref_u, self.selection_alignment,
                                       verbose=True, in_memory=inMemory, weights="mass", strict=strict).run()

        if center:
            if self.mda_u.trajectory.ts.triclinic_dimensions is not None:
                for ts in self.mda_u.trajectory:
                    protein_center = self.mda_u.select_atoms('protein').center_of_mass(pbc=True)
                    dim = ts.triclinic_dimensions
                    box_center = np.sum(dim, axis=0) / 2
                    self.mda_u.atoms.translate(box_center - protein_center)




    def compute_rmsd(self, superposition=False):
        self.mda_u.trajectory[-1]  # set mobile trajectory to last frame
        self.mda_ref_u.trajectory[0]  # set reference trajectory to first frame
        mobile_ca = self.mda_u.select_atoms(self.selection_rmsd)
        ref_ca = self.mda_ref_u.select_atoms(self.selection_rmsd)
        self.rmsd = rms.rmsd(mobile_ca.positions, ref_ca.positions, superposition=superposition)

    def find_shift(self, box_size=None, offset=0):
        """
        Find how much to shift each coordinate to recenter the
            trajectory in the centrer of the box
        """

        # compute center of mass for each frame of the trajectory
        com = np.zeros((self.mda_u.trajectory.n_frames, 3))
        coords = np.zeros((self.mda_u.trajectory.n_frames, self.mda_u.select_atoms('protein').atoms.n_atoms, 3))
        count = 0
        for ts in self.mda_u.trajectory:
            com[count] = np.asarray(self.mda_u.select_atoms('protein').center_of_mass())
            coords[count] = (np.asarray(self.mda_u.select_atoms('protein').center_of_mass()))
            count += 1

        com = np.asarray(com)
        # average com
        avg_com = com.mean(axis=0)

        # round to nearest multiple of 5
        avg_com = [5 * round(c/5) for c in avg_com]
        print('@>: avg. center of mass:', avg_com)

        # find box dimensions
        if box_size is None:
            minx = np.min(coords[:, :, 0]-offset)
            miny = np.min(coords[:, :, 1]-offset)
            minz = np.min(coords[:, :, 2]-offset)

            maxx = np.max(coords[:, :, 0]+offset)
            maxy = np.max(coords[:, :, 1]+offset)
            maxz = np.max(coords[:, :, 2]+offset)

            a = np.array([minx, miny, minz])
            b = np.array([maxx, maxy, maxz])
            # print('@>: box dimensions:', b-a)
            lenx = np.linalg.norm(b[0] - a[0])
            leny = np.linalg.norm(b[1] - a[1])
            lenz = np.linalg.norm(b[2] - a[2])

            # round box length to nearest multiple of 5
            box_lenx = float(5 * np.round(lenx/5))
            box_leny = float(5 * np.round(leny/5))
            box_lenz = float(5 * np.round(lenz/5))

            self.box = [box_lenx, box_leny, box_lenz]
        else:
            self.box = box_size

        print('@>: box dimensions:', self.box)

        # center of box
        center_of_box = [self.box[0]/2, self.box[1]/2, self.box[2]/2]
        shift_x = -np.round(avg_com[0] - center_of_box[0])
        shift_y = -np.round(avg_com[1] - center_of_box[1])
        shift_z = -np.round(avg_com[2] - center_of_box[2])
        self.translate = [shift_x, shift_y, shift_z]


class MDTUniverse:
    def __init__(self):

        self.traj = None
        self.topo = None

        self.traj_ref = None
        self.topo_ref = None

        self.mdt_u = None
        self.mdt_ref_u = None
        self.selection = None               # select indeces of a subset of the topology of interest
        self.selection_alignment = None     # the indices of the atoms to superpose (mobile)
        self.selection_alignment_ref = None # the indices of the atoms to superpose (reference)
        self.use_frame = 0                  # the index of the conformation in reference to align to, if set_use_frame is not called, defalult is zero
        self.rmsd = None
        self.name = None

        self.segid_map = None
        self.chainid_map = None
        self.aligned = None
        self.dihedrals = None

        self.distances_cache = None
        self.distances = None

        self.box = None
        self.translate = None
        self.log = None


    def set_logfile(self, log):
        self.log = log
        logFile = open(self.log, 'w+')
        logFile.close()

    def set_name(self, name):
        self.name = name

    def set_use_frame(self, frame):
        self.use_frame = frame

    def load_universe(self, topo, traj, topo_ref, traj_ref):
        self.mdt_u = md.load(traj, top=topo)
        self.mdt_ref_u = md.load(traj_ref, top=topo_ref)

        print('@>: trajectory contains {} atoms: '.format(self.mdt_u.n_atoms))
        print('@>: reduce trajectory according to following selection: ', self.selection)
        if self.selection != 'all':
            try:
                self.mdt_u.topology.select(self.selection)
                self.mdt_u, self.mdt_ref_u  = tk.reduce_trajectory()  # Reduce trajectory to selected atoms
            except:
                logFile = open(self.log, 'w+')
                tk.print_screen_logfile("@>: selection string not valid -- default ('all') will be used'", logFile)
                self.selection = 'all'

    def source_universe(self, traj, reference=None):
        if reference is not None:
            self.mdt_ref_u = reference
        else:
            self.mdt_u = traj

    def set_selection(self, group_selection):
        self.selection = group_selection

    def set_selection_alignment_ref(self, atm_selection):
        self.selection_alignment_ref = atm_selection

    def set_selection_alignment(self, atm_selection):
        self.selection_alignment = atm_selection

    def set_mapping_segid(self, mapping):
        # Example mapping=lambda X: 0 if X <= 253 else 1
        self.segid_map = mapping

    def set_mapping_chainid(self, mapping):
        # example mapping={0: 'A', 1: 'B'}
        self.chainid_map = mapping

    def set_distances_cache(self, dist_cache):
        self.distances_cache = dist_cache

    def align_md_traj(self):
        """
        Align MDTraj Trajectory

        Parameters
        ----------

        self.mdt_u: mobile mdtraj universe
        self.mdt_ref_u: reference mdtraj universe
        self.selection_string: selection string for alignment

        :return self.aligned
        """

        traj = self.mdt_u
        traj_ref = self.mdt_ref_u
        selection = self.selection_alignment
        selection_ref = self.selection_alignment

        print('@>: selection for alignment: ', self.selection_alignment)

        atom_indices = traj.topology.select(selection)
        atom_indices_ref = traj_ref.topology.select(selection_ref)

        print('@>: use frame {} for alignment'.format(self.use_frame))

        self.aligned = traj.superpose(traj_ref, frame=self.use_frame, atom_indices=atom_indices, ref_atom_indices=atom_indices_ref,
                                    parallel=True)
        self.mdt_u.Trajectory = self.aligned # in memory update trajectory with aligned version

    def adjust_topology(self):
        aligned = self.aligned.atom_slice(self.aligned.topology.select(self.selection))
        table, bonds = aligned.topology.to_dataframe()
        if self.chainid_map is None:
            pass
        else:
            table['chainID'] = table['resSeq'].map(self.chainid_map)

        if self.segid_map is None:
            pass
        else:
            table['segmentID'] = table['chainID'].map(self.segid_map)

        aligned.topology = md.Topology.from_dataframe(table, bonds)
        self.aligned = aligned

    def compute_angles(self):
        traj = self.aligned
        phi_idces, phi = md.compute_phi(traj, periodic=False, opt=True)
        psi_idces, psi = md.compute_psi(traj, periodic=False, opt=True)
        stack = np.hstack([phi, psi])
        self.dihedrals = stack

    def compute_dihed(self):
        if self.aligned is None:
            print('@>: using non-aligned trajectory')
            traj = self.mdt_u
        else:
            traj = self.aligned
        indices, dihedrals = md.compute_phi(traj)
        indices_psi, dihedrals_psi = md.compute_psi(traj)
        sines = np.sin(dihedrals)
        cosines = np.cos(dihedrals)
        sines_psi = np.sin(dihedrals_psi)
        cosines_psi = np.cos(dihedrals_psi)
        q = np.zeros((sines.shape[0], sines.shape[1] * 4))
        q[:, ::4] = sines
        q[:, 1::4] = cosines
        q[:, 2::4] = sines_psi
        q[:, 3::4] = cosines_psi
        self.dihedrals = q

    def compute_distances(self):
        if self.aligned is None:
            print('@>: using non-aligned trajectory')
            traj = self.mdt_u

        else:
            traj = self.aligned
        cached_distances = os.path.exists(self.distances_cache)
        if not cached_distances:
            distances = np.empty((traj.n_frames, traj.n_frames), dtype=np.float32)
            for i in trange(traj.n_frames):
                distances[i] = md.rmsd(traj, traj, i)
            self.distances = distances

            pkl_dump(distances, self.distances_cache)
        else:
            with open(self.distances_cache, 'rb') as handle:
                unserialized_data = pkl.load(handle)
                distances = unserialized_data
                handle.close()
            print('@>: load distances from cache')
            self.distances = distances
        return distances

    @jit(nopython=True, parallel=False, cache=True, nogil=True)
    def calc_rmsd_2frames(self, ref, frame):
        """
        RMSD calculation between a reference and a frame.
        This function is "jitted" for better performances
        """
        dist = np.zeros(len(frame), dtype=np.float32)
        for atom in range(len(frame)):
            dist[atom] = ((ref[atom][0] - frame[atom][0]) ** 2 +
                          (ref[atom][1] - frame[atom][1]) ** 2 +
                          (ref[atom][2] - frame[atom][2]) ** 2)
        return np.sqrt(dist.mean())

    def find_shift(self, box_size=None, offset=0):
        """
        Find how much to shift each coordinate to recenter the
        trajectory in the centrer of the box
        """

        if self.aligned is None:
            print('@>: using non-aligned trajectory')
            trajectory = self.mdt_u
        else:
            trajectory = self.aligned
        # compute center of mass for each frame of the trajectory
        com = md.compute_center_of_mass(trajectory)
        # average com
        avg_com = com.mean(axis=0)*10

        # round to nearest multiple of 5
        avg_com = [5 * round(c/5) for c in avg_com]
        print('@>: avg. center of mass:', avg_com)

        # find box dimensions
        if box_size is None:
            minx = np.min(trajectory.xyz[:, :, 0])*10-offset
            miny = np.min(trajectory.xyz[:, :, 1])*10-offset
            minz = np.min(trajectory.xyz[:, :, 2])*10-offset

            maxx = np.max(trajectory.xyz[:, :, 0])*10+offset
            maxy = np.max(trajectory.xyz[:, :, 1])*10+offset
            maxz = np.max(trajectory.xyz[:, :, 2])*10+offset

            a = np.array([minx, miny, minz])
            b = np.array([maxx, maxy, maxz])
            lenx = np.linalg.norm(a[0] - b[0])
            leny = np.linalg.norm(a[1] - b[1])
            lenz = np.linalg.norm(a[2] - b[2])

            # round box length to nearest multiple of 5
            box_lenx = float(5 * np.round(lenx/5))
            box_leny = float(5 * np.round(leny/5))
            box_lenz = float(5 * np.round(lenz/5))

            self.box = [box_lenx, box_leny, box_lenz]
        else:
            self.box = box_size

        print('@>: box dimensions:', self.box)

        # center of box
        center_of_box = [self.box[0]/2, self.box[1]/2, self.box[2]/2]
        shift_x = -np.round(avg_com[0] - center_of_box[0])
        shift_y = -np.round(avg_com[1] - center_of_box[1])
        shift_z = -np.round(avg_com[2] - center_of_box[2])
        self.translate = [shift_x, shift_y, shift_z]

    def centroid(self):
        """
        Compute the centroid of a distance matrix
        """
        std_dev = np.nanstd(self.distances)
        beta = 1
        cluster_centroid = np.exp(-beta * self.distances / std_dev).sum(axis=1).argmax()
        return cluster_centroid


import numpy as np
import MDAnalysis as mda
from MDAnalysis.analysis import align as mda_align
from sklearn.decomposition import PCA

class DimensionalityReduction:

    def __init__(self):

        self.mda_u = None
        self.projected_u = None
        self.atomgroup = None
        self.reference_u = None
        self.selection = None
        self.data = None
        self.data2D = None
        self.transformed = None
        self.projected = None
        self.variance = None
        self.p_components = None
        self.eigenvalues = None
        self.eigenvectors = None
        self.mean = None
        self.natoms = None
        self.nframes = None
        self.spatial_dim = None
        self.explained_variance_ratio = None
        self._dimensionality_reduction = False


    def source_system(self, align=True, **kwargs):
        self.mda_u = kwargs.get('universe')
        self.reference_u = kwargs.get('reference')
        self.selection = kwargs.get('selection')


        if kwargs.__contains__('start'):
            start = kwargs.get('start')
            start = int(start)
        else:
            start = 0
        if kwargs.__contains__('stop'):
            stop = kwargs.get('stop')
            stop = int(stop)
        else:
            stop = -1
        if kwargs.__contains__('step'):
            step = kwargs.get('step')
            step = int(step)
        else:
            step = 1

        if align:
            universe = self.mda_u
            atomgroup = universe.select_atoms(self.selection)
            # prealign to first frame
            prealigner = mda_align.AlignTraj(self.mda_u, self.mda_u, select="backbone", in_memory=True).run()
            # get reference coordinates
            reference_coordinates = self.mda_u.trajectory.timeseries(asel=atomgroup).mean(axis=1)
            self.reference_u = mda.Merge(atomgroup).load_new(reference_coordinates[None, :, :], order="fac")
            aligner = mda_align.AlignTraj(self.mda_u, self.reference_u, select='backbone', in_memory=True).run()

        # Get the indices of the atoms that match the selection.
        atom_group = self.mda_u.select_atoms(self.selection)
        idices_selection = atom_group.indices

        # Get the flattened coordinates of each frame.
        coordinates = []
        for frame in self.mda_u.trajectory:
            coordinates.append(frame.positions[idices_selection].ravel())
        coordinates = np.array(coordinates)

        self.natoms, self.nframes, self.spatial_dim = self.mda_u.trajectory.timeseries(asel=atom_group).shape
        self.data2D = coordinates
        self.data = coordinates.reshape((self.nframes, self.natoms, self.spatial_dim))
        self.mean = np.mean(self.data, axis=0)


    def MVD(self, n_components=None):
        """Find directions of Maximum Variance by applying PCA on the trajectory."""
        pca = PCA()
        pca.fit(self.data2D)
        # eigenvalues and eigenvectors
        self.eigenvectors = pca.components_
        self.eigenvalues = pca.explained_variance_
        self.explained_variance_ratio = pca.explained_variance_ratio_

        # variance and cumulated variance and reduced data
        if n_components is None:
            n = len(self.eigenvalues)
            self.variance = self.eigenvalues
            self.p_components = self.eigenvectors

        else:
            n = n_components
            self.variance = self.eigenvalues[:n]
            self.p_components = self.eigenvectors[:n, :]

        self._dimensionality_reduction = True

        pca.n_components = n

        self.transformed = pca.transform(self.data2D)



    def transformation(self, universe, components=None):
        """
        Transformation projects the data into the new space defined by the principal components.
        """

        if self._dimensionality_reduction is False:
            print('@>: Perform dimensionality reduction first!')
            return

        pcs = self.eigenvectors[:components, :]
        trans = self.transformed[:, :components]
        projected = np.outer(trans, pcs) + self.mean.flatten()

        self.projected = projected.reshape((len(trans), -1, self.spatial_dim))
        self.projected_u = universe.load_new(self.projected, order="fac")


    def to_file(self, **kwargs):
        """
        Writes the projected universe to a file, parameters are passed to the function through **kwargs, as follows:

        Parameters
        ----------
        kwargs: dict,
            example: ``kwargs = {'universe' : MDAanalysis.universe,
                                 'selection': 'protein and name C CA N',
                                 'outdir'   : './',
                                 'outname'  : 'projected_traj'}``
        """

        if kwargs.__contains__('universe'):
            universe = kwargs['universe']

        elif self.mda_u is not None:
            universe = self.mda_u

        else:
            topo = None
            traj = None
            if kwargs.__contains__('topology_file') and kwargs.__contains__('trajectory_file'):
                topo = kwargs['topology_file']
                traj = kwargs['trajectory_file']

            else:
                print('@>: Provide a topology and trajectory keys in parameters dictionary or MDAnalysis Universe')

            if kwargs.__contains__('selection'):
                selection = kwargs['selection']
            else:
                selection = 'protein and name C CA N'

            universe = mda.Universe(topo, traj)

        if kwargs.__contains__('outdir'):
            outdir = kwargs['outdir']
        else:
            outdir = './'

        if kwargs.__contains__('outname'):
            outname = kwargs['outname']
        else:
            outname = 'projected_traj'

        if kwargs.__contains__('start'):
            start = kwargs.get('start')
            start = int(start)
        else:
            start = 0
        if kwargs.__contains__('stop'):
            stop = kwargs.get('stop')
            stop = int(stop)
        else:
            stop = -1
        if kwargs.__contains__('step'):
            step = kwargs.get('step')
            step = int(step)
        else:
            step = 1
        
        self.projected_u.trajectory[0]

        with mda.Writer(outdir + outname + ".xtc", self.projected_u.atoms.n_atoms) as W:
            for ts in self.projected_u.trajectory[start:stop:step]:
                W.write(self.projected_u.atoms)
        W.close()
        self.projected_u.trajectory[0]
        if kwargs.__contains__('savePDB'):
            with mda.Writer(outdir + outname + "_traj.pdb", self.projected_u.atoms.n_atoms) as W:
                for ts in self.projected_u.trajectory[start:stop:step]:
                    W.write(self.projected_u.atoms)
        W.close()
        self.projected_u.trajectory[0]
        with mda.Writer(outdir + outname + ".pdb", self.projected_u.atoms.n_atoms) as W:
            W.write(self.projected_u.atoms)
        W.close()

        return self.projected_u




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# @author: fmaschietto, bcallen95

from ..imports import *


def plot_pca_space(pcfit, labelsarr, names_list, out, **kwargs):
    """
    Plot pca space

    Parameters
    ----------
    pcfit: np.ndarray,
        array of fitted transformed data from [pca_call = sklearn.PCA() --> pcfit = pca_call.fit_transform(data)]
    labelsarr: np.ndarray,
        array of labels
            example: if the data is the concatenation of two simulations then
            ``data = np.concatenate([sim_1, sim_2])`` and ``labelsarr = np.asarray(names_list[0]*len(sim_2) + names_list[1]*len(sim_2))``;
            the order has to match the data.
    names_list: list of strings
        list containing the title name of each simulation
    out: str,
        plot name (``'/path/to/plot/plotname.pdf'``)
    """

    title  = ''
    levels = 20
    colors = sns.color_palette("bright", len(names_list))
    print(len(colors))
    if kwargs:
        if kwargs.__contains__('title'):
            title=kwargs['title']
        else:
            title = 'Simulation'
        if kwargs.__contains__('levels'):
            levels = kwargs['levels']
        else:
            levels = 20

        if kwargs.__contains__('colors'):
            colors= kwargs['colors']
        else:
            colors = sns.color_palette("bright", len(names_list))
    plt.figure(figsize=(20,20))

    s0 = pcfit[:,0]
    s1 = pcfit[:,1]
    df_pca = pd.DataFrame({'PC1': s0, 'PC2': s1, title: labelsarr})

    sns.displot(df_pca, x="PC1", y="PC2", hue=title, kind="kde", palette=colors, levels=levels)

    for label, name, color in zip(pd.unique(labelsarr), names_list, colors):
        ix = np.where(labelsarr == label)[0]
        plt.gca().scatter(s0[ix[0]], s1[ix[0]],  color=color,  marker='>',  s=200, zorder=100, edgecolor='k')
        plt.gca().scatter(s0[ix[-1]],s1[ix[-1]], color=color,  marker='8',  s=200, zorder=100, edgecolor='k')

    m1 = mlines.Line2D([], [], color='w', marker='>', markeredgewidth=.08, markeredgecolor='k',
                              markersize=130, label='Start of simulation')
    m2 = mlines.Line2D([], [], color='w', marker='8',  markeredgewidth=.08, markeredgecolor='k',
                              markersize=130, label='End of simulation')
    plt.savefig(out, format='pdf', bbox_inches='tight')




"""

MDiGest v.0.1.7

__version__ = 0.1.7
__author__ = Federica Maschietto <federica.maschietto@gmail.com>, Brandon Allen <bcallen95@gmail.com>


    DESCRIPTION
    # imports             --> general imports
    # parsetrajectory     --> process trajectory
    # correlation         --> compute correlation based on atomic displacements
    # dcorrelation        --> compute correlation from dihedrals fluctuations
    # kscorrelation       --> KS analysis
    # dimreduction        --> dimensionality reduction
    # community           --> GN, LOUVAIN, communities in general
    # savedata            --> caches the output of various models
    # auxiliary           --> auxiliary functions used by multiple modules
    # toolkit             --> accessory functions
    # plots

"""



from . import imports
from . import parsetrajectory
from . import correlation
from . import dcorrelation
from . import kscorrelation
from . import networkcommunities
from . import savedata
from . import networkcanvas
#from . import dimreduction
from . import auxiliary
from . import toolkit
from . import plots


__all__ = ['parsetrajectory', 'analysis', 'correlation', 'dcorrelation',
           'kscorrelation', 'networkcommunities', 'dimreduction',
           'savedata', 'networkcanvas', 'auxiliary', 'toolkit', 'plots']






"""#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# @author: fmaschietto, bcallen95"""

import pandas           as pd
import numpy            as np
import networkx         as nx
import mdtraj           as md
import MDAnalysis       as mda
import pickle
import h5py
import scipy
import os
import sklearn
from itertools import combinations_with_replacement as cwr
from tqdm import tqdm
from tqdm.notebook import trange
from collections import OrderedDict, defaultdict

import matplotlib.pyplot as plt
import matplotlib as mpl
import matplotlib.lines as mlines
import matplotlib.cm as cm
import seaborn as sns
import warnings

# matplotlib parameters
plt.rcParams['axes.labelpad'] = 3.
plt.rcParams['font.size'] = 30
plt.rcParams['savefig.dpi'] = 300
plt.rcParams['axes.labelsize'] = 25
plt.rcParams['axes.titlesize'] = 25
plt.rcParams['lines.linewidth'] = 2.0
plt.rcParams['lines.markersize'] = 10
plt.rcParams['legend.fontsize'] = 25
plt.rcParams['text.usetex'] = False
plt.rcParams['font.family'] = "Dejavu Sans"
plt.rcParams['font.serif'] = "Dejavu Sans"
plt.rcParams['figure.autolayout'] = True




"""#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# @author: fmaschietto, bcallen95"""

from   mdigest.core.imports import *
import mdigest.core.toolkit as tk
import mdigest.core.savedata as sd
import mdigest.core.auxiliary as aux

from operator                 import itemgetter as ig
from sklearn.preprocessing    import StandardScaler
from MDAnalysis.lib.distances import capped_distance

class DynCorr:
    """General purpose class handling computation of different correlation metrics from atomic displacements sampled over MD trajectories."""

    def __init__(self, MDSIM):
        """
        Description
        -----------
        General purpose class handling computation of different correlation metrics from atomic displacements sampled over MD trajectories.

        Parameters
        ----------
        MDSIM: class object
        DynCorr inherits general attributes from MDSIM


        Methods
        -------


        Attributes
        ----------

        self.coordinates_allreplicas: dict, with replica index ``rep_n`` as key and values of shape (nsamples, nfeatures * features_dimension)
            coordinates array.

        self.displacements_allreplicas: dict, with replica index ``rep_n`` as key and values of shape shape (nsamples, nfeatures * features_dimension)
            array containing the displacement of each residue from the average position (the mean x, y, z coordinates
            over all the selected timesteps)

        self.distances_allreplicas: dict, with replica index ``rep_n`` as key and values of shape (nfeatures, nfeatures)
            node to node pairwise distances computed from average positions for each given trajectory replica

        self.disp_from_mean_allreplicas: dict, with replica index ``rep_n`` as key and values of shape (nsamples, nfeatures)
            array containing the displacement of each atom from the average position computed over all the selected timesteps

        self.covar_disp_allreplicas: dict, with replica index ``rep_n`` as key and values of shape (nfeatures, nfeatures)
            covariance matrix of atomic displacements for each given trajectory replica

        self.gcc_allreplicas: nested dict, with replica index ``rep_n`` as key and inner dict with key ``gcc_mi`` or ``gcc_lmi`` and values of shape (nfeatures, nfeatures)
            linearized Mutual Information based generalized correlation coefficient matrix[1]. Each entry represents the pairwise linearized generalized correlation coefficient between each pair of nodes,
            for each given trajectory replica. ``gcc_lmi`` linearized mutual information based generalized correlation using gaussian estimator; ``gcc_mi`` mutual information based generalized correlation computed using nonlinear estimator.

        self.dcc_allreplicas: dict, with replica index ``rep_n`` as key and values of shape (nfeatures, nfeatures)
            normalized dynamical cross-correlations matrix

        self.pcc_allreplicas: dict, with replica index ``rep_n`` as key and values of shape (nfeatures, nfeatures)
            Pearson's product-moment correlation coefficients.

        self.exclusion_matrix_allreplicas: dict, with replica index ``rep_n`` as key and values of shape (nfeatures, nfeatures)
            which nodes to consider for calculation of communities

        self.eigenvector_centrality_allreplicas: dict, with replica index ``rep_n`` as key and values of shape (nfeatures)
            eigenvector centrality arrays for each given trajectory replica


        References
        ----------
        [1] Lange, O.; Grubmueller, H.: Generalized correlation for biomolecular dynamics.
        Proteins: Structure, Function and Bioinformatics 62 (4), pp. 1053 - 1061 (2006)


        Examples
        --------
        """
        self.mds_data                = MDSIM.mds_data
        self.mda_u                   = MDSIM.mda_u
        self.atom_group_selstr       = MDSIM.atom_group_selstr
        self.system_selstr           = MDSIM.system_selstr
        self.atom_group_selection    = MDSIM.atom_group_selection
        self.nodes_idx_array         = MDSIM.nodes_idx_array
        self.nodes_to_res_dictionary = MDSIM.nodes_to_res_dictionary

        # ---------------------------------------------#

        self.coordinates_allreplicas            = {}
        self.displacements_allreplicas          = {}
        self.distances_allreplicas              = {}
        self.disp_from_mean_allreplicas         = {}
        self.covar_disp_allreplicas             = {}
        self.gcc_allreplicas                    = {}
        self.dcc_allreplicas                    = {}
        self.pcc_allreplicas                    = {}
        self.exclusion_matrix_allreplicas       = {}
        self.eigenvector_centrality_allreplicas = {}

        # ---------------------------------------------#

        self.natoms              = MDSIM.natoms
        self.total_nframes       = MDSIM.total_nframes
        self.nframes_per_replica = MDSIM.nframes_per_replica
        self.num_replicas        = MDSIM.num_replicas
        self.nresidues           = MDSIM.nresidues
        self.nnodes              = MDSIM.nnodes # nfeatures
        self.initial             = MDSIM.initial
        self.final               = MDSIM.final
        self.step                = MDSIM.step
        self.window_span         = MDSIM.window_span # nsamples

    def save_class(self, file_name_root='./output/cache/'):
        """
        can be used to dump all correlation analyses (produced upon calling the correlation, kscorrelation,
        and dcorrelation modules) or the community analysis.

        Parameters
        ----------
        file_name_root: str
            filename rootname
        """

        self.mds_data = sd.MDSdata()
        self.mds_data.atom_group_selstr                  = self.atom_group_selstr
        self.mds_data.system_selstr                      = self.system_selstr
        self.mds_data.nodes_idx_array                    = self.atom_group_selection.ix_array
        self.mds_data.nodes_to_res_dictionary            = self.nodes_to_res_dictionary # useful for later plotting (match nodes id to resid)
        self.mds_data.nresidues                          = self.nresidues
        self.mds_data.nnodes                             = self.nnodes
        self.mds_data.num_replicas                       = self.num_replicas
        self.mds_data.disp_from_mean_allreplicas         = self.disp_from_mean_allreplicas
        self.mds_data.covar_disp_allreplicas             = self.covar_disp_allreplicas
        self.mds_data.coordinates_allreplicas            = self.coordinates_allreplicas
        self.mds_data.displacements_allreplicas          = self.displacements_allreplicas
        self.mds_data.distances_allreplicas              = self.distances_allreplicas
        self.mds_data.gcc_allreplicas                    = self.gcc_allreplicas
        self.mds_data.dcc_allreplicas                    = self.dcc_allreplicas
        self.mds_data.pcc_allreplicas                    = self.pcc_allreplicas
        self.mds_data.exclusion_matrix_allreplicas       = self.exclusion_matrix_allreplicas
        self.mds_data.eigenvector_centrality_allreplicas = self.eigenvector_centrality_allreplicas
        self.mds_data.save_to_file(file_name_root)


    def edge_exclusion(self, spatial_cutoff=4.5, contact_cutoff=.75, save_name='none'):
        # TODO priority 1 MOVE THIS TO COMMUNITIES MODULE
        # TODO change nresidues with nnodes.
        # Warning: this function was only tested with the option of number of nodes in the calculation chosen equal to
        # nresidues

        """
        Compute the so-called `edge exclusion matrix`. When analyzing correlations it can be useful to analyze only correlation corresponding to residue pairs that
        are proximal (below a certain distance threshold) for a certain percentage of the trajectory frames. This function takes care of computing this using `capped_distance` from MDAnalysis.

        Parameters
        ----------
        spatial_cutoff: float,
            distance threshold to define atoms in contact (in Amstrong)
        contact_cutoff: float,
            contact persistency above which to consider pair expressed in percentage of frames
        save_name:
            output filename
        """

        def _multiplicity_pairs(pair_array, threshold):
            x = pair_array.transpose()[0]
            y = pair_array.transpose()[1]
            m = np.max(pair_array) * 10.0
            h = x * m + y
            unique, count = np.unique(h, return_counts=True)
            good_pairs_h = unique[count >= threshold]
            x_g = (np.floor(good_pairs_h / float(m))).astype(int)
            y_g = (good_pairs_h % m).astype(int)
            good_pairs = np.transpose(np.vstack([x_g, y_g]))
            return good_pairs

        def _unique_pairs(pair_array):
            x = pair_array.transpose()[0]
            y = pair_array.transpose()[1]
            m = 1000000
            h = x * m + y
            unique, count = np.unique(h, return_counts=True)
            x_g = (np.floor(unique / float(m))).astype(int)
            y_g = (unique % m).astype(int)
            return np.transpose(np.vstack([x_g, y_g]))

        def _compute_exclusion_matrix(universe, sys_sele_str, nres, i_frame, f_frame, n_space, spatialcutoff,
                                      contactcutoff, savename='None'):
            """
            Computation of exclusion matrix. The function is fed with instances from MDSIM class / arguments at provided at call
            TODO cleanup, convert function into class method

            Parameters
            ----------
            universe: object,
                MDA universe
            sys_sele_str:    str,
                defining the subset of atoms used in the computation of the exclusion matrix
            nres:            int,
                number of residues (nodes)
            i_frame:         int,
                initial frame of the trajectory used for exclusion matrix calculation
            f_frame:         int,
                final frame of the trajectory used for exclusion matrix calculation
            n_space:         int,
                trajectory step used for exclusion matrix calculation
            spatialcutoff:   float,
                distance threshold defining whether two atoms (nodes) are in contact
            contactcutoff:   float,
                contact persistency threshold defining whether two atoms (nodes) are in contact
            savename:        str,
                output filename

            Returns
            -------
            excl_mat: np.ndarray
                exclusion matrix
            """

            if savename != 'None':
                print('@>: saving exclusion matrixt to ', savename)
                m = tk.retrieve(savename + ".eemat")

                if m is not None:
                    return m

            residues = universe.select_atoms(sys_sele_str, updating=True).residues

            natoms = len(universe.select_atoms(sys_sele_str, updating=True).residues.atoms)

            print('@>: number of selected atoms: ',
                  len(universe.select_atoms(sys_sele_str, updating=True).residues.atoms.ids))

            atom_to_res_dictionary = dict(zip(universe.select_atoms(sys_sele_str).atoms.ids,
                                              universe.select_atoms(sys_sele_str).residues.atoms.resindices))

            residue_name_to_index = {}
            for indx in range(len(residues)):
                residue_name_to_index[residues[indx].resindex] = indx

            inclusion_matrix = np.zeros((nres, nres))
            mega_list = []
            noth_ids = residues.atoms.select_atoms("not type H*").ids

            print("---------------------------------------------------------------")
            print("@>: computing exclusion matrix...")
            print("@>: first:  %d, last: %d, step: %d" % (i_frame, f_frame, n_space))
            print("---------------------------------------------------------------")

            trajectory = universe.trajectory[i_frame:f_frame:n_space]
            trajlen = len(trajectory)
            for ts in trajectory:
                not_h = residues.atoms.select_atoms("not type H*", updating=True)

                pairs = capped_distance(not_h.positions, not_h.positions, spatialcutoff, return_distances=False)

                second_atoms = np.transpose(pairs)[1]
                first_atoms = np.transpose(pairs)[0]

                second_atoms = noth_ids[second_atoms]
                first_atoms = noth_ids[first_atoms]

                a2r_selection1 = np.asarray(ig(*first_atoms)(atom_to_res_dictionary))
                a2r_selection2 = np.asarray(ig(*second_atoms)(atom_to_res_dictionary))

                res_pairs = np.transpose(np.vstack([a2r_selection1, a2r_selection2]))
                unique_res_pairs = _unique_pairs(res_pairs)
                mega_list.append(unique_res_pairs)

            mega_list = np.vstack(mega_list)
            good_atom_pairs = _multiplicity_pairs(mega_list, contactcutoff * trajlen)

            for pair in good_atom_pairs:
                res_1 = residue_name_to_index[pair[0]]
                res_2 = residue_name_to_index[pair[1]]
                inclusion_matrix[res_1, res_2] = 1
                inclusion_matrix[res_2, res_1] = 1

            excl_mat = 1 - inclusion_matrix
            for i in range(0, nres - 1):
                excl_mat[i, i + 1] = 1.0

            for i in range(1, nres):
                excl_mat[i, i - 1] = 1.0

            np.fill_diagonal(excl_mat, 1.0)

            if savename != 'None':
                out_name = save_name + '_%d' % win_idx + ".eemat"
                tk.dump(out_name, excl_mat)

            print("---------------------------------------------------------------")
            print("@>: done computing exclusion matrix.")
            print("@>: found ", (nres * nres) - np.sum(excl_mat), 'nonzero elements in the exclusion matrix')
            print("---------------------------------------------------------------")
            return excl_mat

        exclusion_matrix_allreplicas = {}

        for win_idx in tk.log_progress(range(self.num_replicas), every=1, size=self.num_replicas, name="Window"):
            #beg = int(self.final / self.num_replicas) * win_idx
            #end = int(self.final / self.num_replicas) * (win_idx + 1)

            beg = self.initial + self.window_span * win_idx
            end = self.initial + self.window_span * (win_idx + 1)


            print('@>: using frames %d to %d with step %s' % (beg, end, self.step))

            name = save_name + "_%d" % win_idx

            exclusion_matrix_allreplicas['rep_%d' %win_idx] = _compute_exclusion_matrix(self.mda_u, self.system_selstr,
                                                                              self.nresidues,
                                                                              beg, end, self.step, spatial_cutoff,
                                                                              contact_cutoff, name)

        self.exclusion_matrix_allreplicas = exclusion_matrix_allreplicas


    def parse_dynamics(self, scale=False, normalize=True, LMI='gaussian', MI='knn_5_1', DCC=False, PCC=False, COV_DISP=False,
                       CENTRALITY=True, VERBOSE=False, **kwargs):
        """
        Parse molecular dynamics trajectory and compute different correlation metrices

        Parameters
        ----------
        scale: bool,
            whether to remove mean from coordinates using StandardScaler

        normalize: bool,
            whether to normalize cross-correlation matrices

        LMI: str or None,
            - None to skip computation of LMI based correlation
            - 'gaussian' to compute LMI

        MI: str,
            -None to skip computation of MI based correlation
            -'knn_arg1_arg2' to compute MI, with k = arg1, and estimator= arg2, default is 'knn_5_1'

        DCC: bool,
            whether to compute dynamical cross-correlation matrix of atomic displacements. Default is False

        PCC: bool,
            whether to compute Pearson correlation matrix of atomic displacements. Default is False

        COV_DISP: bool,
            whether to compute the covariance of atomic displacements. Default is False

        CENTRALITY: bool,
            whether to compute centrality of atomic displacements. Default is True

        VERBOSE: bool,
            whether to set verbose printing

        """

        # Assign stride variable
        stride = self.step

        if self.nresidues != self.nnodes:
            nr = self.nnodes
            print('@>: sanity check warning: number of nodes included exceeds the number of residues')
        else:
            nr = self.nresidues
            print('@>: sanity check pass: number of residues is same as number of nodes')

        # Number or replicas
        num_replicas = self.num_replicas

        # For 3D atom position data
        feat_dimension = 3

        # Stores all data in a frame-by-dimension format.
        coordinates                = np.zeros((num_replicas, self.window_span, nr, feat_dimension),  dtype=float)
        coordinates_allreplicas    = np.zeros((num_replicas, self.window_span, nr * feat_dimension), dtype=float)
        displacements_allreplicas  = np.zeros((num_replicas, self.window_span, nr * feat_dimension))
        disp_from_mean_allreplicas = np.zeros((num_replicas, self.window_span, nr))
        covar_disp_allreplicas     = np.zeros((num_replicas, nr, nr))
        distances_allreplicas      = np.zeros((num_replicas, nr, nr))
        dcc_allreplicas            = np.zeros((num_replicas, nr, nr))
        pcc_allreplicas            = np.zeros((num_replicas, nr, nr))

        if VERBOSE:
            print('@> -- shape of coordinates matrix:',             coordinates.shape)
            print('@> -- shape of coordinates_allreplicas matrix:', coordinates_allreplicas.shape)
            print('@> -- shape of displacements_allreplicas matrix:', displacements_allreplicas.shape)
            print('@> -- shape of disp_from_mean_allreplicas matrix:', disp_from_mean_allreplicas.shape)
            print('@> -- shape of distances_allreplicas matrix:', distances_allreplicas.shape)
            print('@> -- shape of covar_disp_allreplicas', covar_disp_allreplicas.shape)
            print('@> -- shape of dcc_allreplicas matrix:', dcc_allreplicas.shape)
            print('@> -- shape of pcc_allreplicas matrix:', pcc_allreplicas.shape)


        # setup kwargs
        try:
            subset = kwargs['subset']
        except KeyError:
            subset = None

        solvers = []
        if MI  is not None:
            solvers.append(MI)
        if LMI is not None:
            solvers.append(LMI)


        for win_idx in tk.log_progress(range(num_replicas), every=1, size=num_replicas, name="Window"):

            beg = self.initial + self.window_span * win_idx
            end = self.initial + self.window_span * (win_idx + 1)

            print("@>: LMI/MI calculation ...")
            print("@>: begin frame: %d" % beg)
            print("@>: end   frame: %d" % end)
            print("@>: step:        %d" % self.step)

            counter = 0
            for frame in self.mda_u.trajectory[beg:end:stride]:
                coordinates[win_idx, counter, :, :] = self.atom_group_selection.positions
                counter += 1
            counter = 0

            if COV_DISP:
                print("@>: compute covariance of displacements...")

                disp_from_mean_allreplicas[win_idx] = aux._compute_square_displacements(coordinates[win_idx, :, :, :])

                if scale:
                    # nfeatures displacements from mean are not centered hence scaling will have an effect
                    tmp = disp_from_mean_allreplicas[win_idx].copy()
                    scaler = StandardScaler(with_std=False).fit(tmp)
                    disp_from_mean_allreplicas[win_idx] = scaler.transform(tmp)

                    covar_disp_allreplicas[win_idx] = np.cov(disp_from_mean_allreplicas[win_idx], rowvar=False, bias=True)

                else:
                    covar_disp_allreplicas[win_idx] = np.cov(disp_from_mean_allreplicas[win_idx], rowvar=False, bias=True)


            print("@>: reshaping coordinates...")

            coordinates_allreplicas[win_idx] = aux.coordinate_reshape(coordinates[win_idx, :, :, :])

            if scale:
                # scaling has no effect because 3*nfeatures displacements are already mean averaged
                c_ = coordinates[win_idx, :, :].copy()
                tmp = aux._compute_displacements(c_)
                scaler = StandardScaler(with_std=False).fit(tmp)
                displacements_allreplicas[win_idx] = scaler.transform(tmp)
            else:
                print("@>: computing the displacement vectors (X- <X>)...")
                displacements_allreplicas[win_idx] = aux._compute_displacements(coordinates[win_idx, :, :])

            MIdict = {}
            ECdict = {}
            if solvers is not []:
                for solver in solvers:
                    if solver == 'gaussian':
                        MIdict.update({'gcc_lmi': aux.compute_generalized_correlation_coefficients(displacements_allreplicas[win_idx].reshape((self.window_span, nr, feat_dimension)),
                                                                                         features_dimension=feat_dimension, solver=solver, correction=False)})
                        if CENTRALITY:
                            print("@>: computing eigenvector centrality from lmi matrix")
                            _, ec = aux.compute_eigenvector_centrality(MIdict['gcc_lmi'], weight='weight')
                            ECdict.update({'gcc_lmi': ec})




                    elif 'knn' in solver:
                        MIdict.update({'gcc_mi': aux.compute_generalized_correlation_coefficients(displacements_allreplicas[win_idx].reshape((self.window_span, nr, feat_dimension)),
                                                                         features_dimension=feat_dimension, solver=solver, correction=True, subset=subset)})
                        if CENTRALITY:
                            print("@>: computing eigenvector centrality from mi matrix")
                            _, ec = aux.compute_eigenvector_centrality(MIdict['gcc_mi'], weight='weight')
                            ECdict.update({'gcc_mi': ec})

            self.gcc_allreplicas['rep_%d' % win_idx] = MIdict
            self.eigenvector_centrality_allreplicas['rep_%d' % win_idx] = ECdict

            print("@>: computing and storing distances...")
            distances_allreplicas[win_idx] = aux.compute_distance_matrix(coordinates_allreplicas[win_idx, :, :], feat_dimension)

            if DCC:
                print("@>: computing normalized dynamical cross-correlation matrix")
                dcc_allreplicas[win_idx] = aux.compute_DCC(coordinates_allreplicas[win_idx], features_dimension=feat_dimension, normalized=normalize)

            if PCC:
                print("@>: computing pearson correlation matrix")
                pcc_allreplicas[win_idx] = np.corrcoef(disp_from_mean_allreplicas[win_idx], rowvar=False)

        # Store as class instances
        self.coordinates_allreplicas     = dict(zip(['rep_%d' % i for i  in range(self.num_replicas)], coordinates_allreplicas))
        self.displacements_allreplicas   = dict(zip(['rep_%d' % i for i  in range(self.num_replicas)], displacements_allreplicas))
        self.distances_allreplicas       = dict(zip(['rep_%d' % i for i  in range(self.num_replicas)], distances_allreplicas))
        self.dcc_allreplicas             = dict(zip(['rep_%d' % i for i  in range(self.num_replicas)], dcc_allreplicas))
        self.pcc_allreplicas             = dict(zip(['rep_%d' % i for i  in range(self.num_replicas)], pcc_allreplicas))
        self.disp_from_mean_allreplicas  = dict(zip(['rep_%d' % i for i  in range(self.num_replicas)], disp_from_mean_allreplicas))
        self.covar_disp_allreplicas      = dict(zip(['rep_%d' % i for i  in range(self.num_replicas)], covar_disp_allreplicas))




"""#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# @author: fmaschietto, bcallen95"""
import numpy as np

from mdigest.core.imports                 import *


class MDS:
    """Parse molecular dynamics trajectories"""
    def __init__(self):
        """
        General purpose class to parse molecular dynamics trajectories, based on MDAnalysis.
        ``[**]`` function structure adapted from https://github.com/melomcr/dynetan

        Attributes
        ----------
        self.mds_data: class object,
            for saving
        self.mda_u: object,
            MDAnalysis universe
        self.atom_group_selstr: str,
            atom group selection string
        self.atom_group_selection: object,
            atom group selected based on atom_group_selstr from mda_u.select_atoms(atom_group_selstr)
        self.nodes_idx_array: np.ndarray,
            indices of selected atoms
        self.nodes_to_res_dictionary: dict
            dictionary with nodes (atoms) indices as keys and residues as values
        self.natoms: int,
            number of atoms in selected atom group
        self.total_nframes: int,
            total number of frames i MDA universe before slicing and trimming
        self.nframes_per_replica: int,
            number of frames per replica
        self.num_replicas: int,
            number of replicas
        self.nresidues: int,
            number of residues in selected atom group
        self.initial: int,
            starting frame read in from trajectory
        self.final: int,
            final frame read in from trajectory
        self. step: int,
            step between each frame
        self.window_span: int,
            window span

        References
        ----------

        Examples
        --------
        """
        self.mds_data                = None
        self.mda_u                   = None
        self.atom_group_selstr       = None
        self.system_selstr           = None
        self.atom_group_selection    = None
        self.nodes_idx_array         = None
        self.nodes_to_res_dictionary = None

        self.natoms              = 0
        self.total_nframes       = 0
        self.nframes_per_replica = 0
        self.num_replicas        = 0
        self.nresidues           = 0
        self.nnodes              = 0
        self.initial             = 0
        self.final               = 0
        self.step                = 0
        self.segment             = 0
        self.window_span         = 0


    def set_num_replicas(self, num_replicas):
        """
        Set the number of replicas

        Parameters
        ----------
        num_replicas: int,
            number of concatenated replicas
        """

        self.num_replicas = num_replicas


    def source_system(self, mda_universe):
        """
        Source MDA universe

        Parameters
        ----------
        mda_universe: mda.Universe object,
            MDA universe object
        """
        self.mda_u = mda_universe.copy()


    def load_system(self, topology, traj_files, inmem=True):
        """
        Load MDA universe from topology and trajectory

        Parameters
        ----------
        topology: str,
            path to topology file
        traj_files: str or list of str,
            strings or list of strings specifying the path to trajectory file/s
        inmem: bool,
            whether to load trajectory in memory
        """
        self.mda_u = mda.Universe(topology,traj_files, in_memory=inmem)


    def align_traj(self, inmem=True, reference=None, selection='protein'):
        """
        Align trajectory to specified selection using aling protocol from MDAnalysis

        Parameters
        ----------
        inmem: bool, default True,

        reference: bool or None, defalult None,
            a reference universe can be specified to use against for alignment
        selection: str,
            selection string to select atoms against which to perform alignment
        """
        from MDAnalysis.analysis import align as mdaAlign

        # Set the first frame as reference for alignment
        self.mda_u.trajectory[0]

        if reference is None:
            alignment = mdaAlign.AlignTraj(self.mda_u, self.mda_u, select="%s and not (name H* or name [123]H*)" % selection,
                                       verbose=True, in_memory=inmem, weights="mass")
            alignment.run()
        else:
            alignment = mdaAlign.AlignTraj(self.mda_u, reference, select="%s and not (name H* or name [123]H*)" % selection,
                                       verbose=True, in_memory=inmem, weights="mass")
            alignment.run()


    def set_selection(self, atom_group_selstr='protein and name CA', system_selstr='protein'):
        """
        Set selection strings

        Parameters
        ----------
        system_selstr: str,
            selection string to select the system portion to consider when computing the exclusion matrix
            for example "protein"
        atom_group_selstr: str,
            selection string to be used for selecting the subset of the nodes on which to perform analysis
            for example "protein and name CA"
        """
        self.atom_group_selstr = atom_group_selstr
        self.system_selstr = system_selstr


    def get_universe(self):
        """
        Retrieve universe
        """
        return self.mda_u


    def stride_trajectory(self, initial=0, final=-1, step=1):
        """
        Stride trajectory

        Parameters
        ----------
        initial: int,
            initial frame from which to start reading in the trajectory
        final: int,
            final frame to consider when reading in the trajectory
        step: int,
            step to use when slicing the traj frames
        """

        # Select atom group of interest
        self.atom_group_selection = self.mda_u.select_atoms(self.atom_group_selstr)

        # Set natoms attribute as the number of atoms in the selected atom group
        self.natoms  = self.atom_group_selection.atoms.n_atoms

        # Set nodes_idx_array attribute from the array of indices of each atom in the selected atom group
        # Each atom is assigned an index (interger between 0 and N with N total number of atoms in the MDUniverse)
        self.nodes_idx_array = self.atom_group_selection.ix_array

        if final < 0:
            final = self.mda_u.trajectory.n_frames
        else:
            final = final

        # Total number of frames in the original trajectory
        self.total_nframes = len(self.mda_u.trajectory)

        # Len of trajectory segment to analyze
        self.segment = self.segment

        # Number or replicas
        self.num_replicas = self.num_replicas

        # Initial frame
        self.initial = initial

        # Final frame
        self.final = final

        # Step
        self.step =  step

        def slice_time_series(time_series, start_frame, last_frame, nstep, number_of_replicas ):
            if last_frame == -1:
                last_frame = len(time_series)
                print(len(time_series))
            segment = time_series[start_frame:last_frame]
            chunk_size = len(segment) // number_of_replicas

            if ((chunk_size / nstep) % 1 == 0.) and ((len(segment) / number_of_replicas) % 1 == 0.):
                chunk_size = len(range(chunk_size)[::nstep])
            else:

                segment = time_series[start_frame:last_frame-1]
                chunk_size = len(range(chunk_size-1)[::nstep])
            return len(segment), chunk_size

        # Length of trajectory segment
        self.segment, self.window_span = slice_time_series(self.mda_u.trajectory,
                                                           self.initial, self.final, self.step, self.num_replicas)

        # Number of frames per replica
        self.nframes_per_replica = self.window_span# int(len(self.mda_u.trajectory[initial:final:step])/self.num_replicas)

        print('@>: total number of frames in trajectory file: %d' % self.total_nframes)
        print('@>: length of selected trajectory segment:     %d' % (int(self.window_span) * int(self.num_replicas)))
        if self.total_nframes != (int(self.window_span) * int(self.step) * int(self.num_replicas)):
            print("@>: WARNING: the segmented trajectory has an uneven number of frames")
            print("@>: WARNING: %d last frame/s will be excluded" % ((self.final - self.initial) -
                                                                     int(self.window_span) * int(self.step) * int(self.num_replicas)))
        print('@>: number of replicas:                     %d' % self.num_replicas)
        print("@>: using window length of %d simulation steps"   % self.window_span)
        print('@>: number or frames per replica:           %d' % self.nframes_per_replica)
        print('@>: first frame:                            %d' % initial)
        print('@>: last frame:                             %d' % final)
        print('@>: actual last frame:                      %d' % self.segment)
        print('@>: step:                                   %d' % step   )

        # Number of residues in selected atom group
        self.nresidues = len(self.atom_group_selection.residues)
        self.nnodes    = len(self.atom_group_selection.resids)
        if self.nnodes != self.nresidues:
            print('@>: warning: selected nodes exceed number of residues')
        print('@>: number of residues in selected atom group: %d' % self.nresidues)
        print('@>: number of nodes    in selected atom group: %d' % self.nnodes)
        print('@>: number of elements in selected atom group: %d' % len(self.atom_group_selection))

        self.nodes_to_res_dictionary = dict(zip(self.mda_u.select_atoms(self.atom_group_selstr).atoms.ids,
                                          self.atom_group_selection.atoms.resindices))

